# Copyright 2013 OpenStack Foundation
# Copyright (c) 2015 Mirantis inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Tests for manila.api.v2.quota_sets.py
"""

import ddt
import mock
from oslo_config import cfg
import webob.exc
import webob.response

from manila.api.openstack import api_version_request as api_version
from manila.api.v2 import quota_sets
from manila import context
from manila import exception
from manila import policy
from manila import test
from manila.tests.api import fakes
from manila import utils

CONF = cfg.CONF


def _get_request(is_admin, user_in_url):
    req = mock.MagicMock(
        api_version_request=api_version.APIVersionRequest("2.40"))
    req.environ = {'manila.context': context.get_admin_context()}
    req.environ['manila.context'].is_admin = is_admin
    req.environ['manila.context'].auth_token = 'foo_auth_token'
    req.environ['manila.context'].project_id = 'foo_project_id'
    if user_in_url:
        req.environ['manila.context'].user_id = 'foo_user_id'
        req.environ['QUERY_STRING'] = 'user_id=foo_user_id'
    return req


@ddt.ddt
class QuotaSetsControllerTest(test.TestCase):

    def setUp(self):
        super(QuotaSetsControllerTest, self).setUp()
        self.controller = quota_sets.QuotaSetsController()
        self.resource_name = self.controller.resource_name
        self.project_id = 'foo_project_id'
        self.mock_policy_check = self.mock_object(
            policy, 'check_policy', mock.Mock(return_value=True))

    @ddt.data(
        {"shares": 3, "snapshots": 4, "gigabytes": 5,
         "snapshot_gigabytes": 6, "share_networks": 7},
        {"shares": -1, "snapshots": -1, "gigabytes": -1,
         "snapshot_gigabytes": -1, "share_networks": -1},
        {"shares": 13},
        {"snapshots": 24},
        {"gigabytes": 7},
        {"snapshot_gigabytes": 10001},
        {"share_networks": 12345},
        {"share_groups": 123456},
        {"share_group_snapshots": 123456},
    )
    def test_defaults(self, quotas):
        req = _get_request(True, False)
        for k, v in quotas.items():
            CONF.set_default('quota_' + k, v)
        expected = {
            'quota_set': {
                'id': self.project_id,
                'shares': quotas.get('shares', 50),
                'gigabytes': quotas.get('gigabytes', 1000),
                'snapshots': quotas.get('snapshots', 50),
                'snapshot_gigabytes': quotas.get('snapshot_gigabytes', 1000),
                'share_networks': quotas.get('share_networks', 10),
                'share_groups': quotas.get('share_groups', 50),
                'share_group_snapshots': quotas.get(
                    'share_group_snapshots', 50),
            }
        }

        result = self.controller.defaults(req, self.project_id)

        self.assertEqual(expected, result)
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'show')

    @ddt.data(
        ('os-', '1.0', quota_sets.QuotaSetsControllerLegacy, 'defaults'),
        ('os-', '2.6', quota_sets.QuotaSetsControllerLegacy, 'defaults'),
        ('', '2.7', quota_sets.QuotaSetsController, 'defaults'),
        ('os-', '1.0', quota_sets.QuotaSetsControllerLegacy, 'show'),
        ('os-', '2.6', quota_sets.QuotaSetsControllerLegacy, 'show'),
        ('', '2.7', quota_sets.QuotaSetsController, 'show'),
    )
    @ddt.unpack
    def test_get_quotas_with_different_api_versions(self, url, version,
                                                    controller, method_name):
        expected = {
            'quota_set': {
                'id': self.project_id,
                'shares': 50,
                'gigabytes': 1000,
                'snapshots': 50,
                'snapshot_gigabytes': 1000,
                'share_networks': 10,
            }
        }
        req = fakes.HTTPRequest.blank(
            '/fooproject/%squota-sets' % url,
            version=version, use_admin_context=True)

        result = getattr(controller(), method_name)(req, self.project_id)

        self.assertEqual(expected, result)

    @staticmethod
    def _get_share_type_request_object(microversion=None):
        req = _get_request(True, False)
        req.environ['QUERY_STRING'] = 'share_type=fake_share_type_name_or_id'
        req.api_version_request = api_version.APIVersionRequest(
            microversion or '2.39')
        return req

    @ddt.data('2.39', '2.40')
    def test_share_type_quota_detail(self, microversion):
        self.mock_object(
            quota_sets.db, 'share_type_get_by_name_or_id',
            mock.Mock(return_value={'id': 'fake_st_id'}))
        req = self._get_share_type_request_object(microversion)
        quotas = {
            "shares": 23,
            "snapshots": 34,
            "gigabytes": 45,
            "snapshot_gigabytes": 56,
        }
        expected = {'quota_set': {
            'id': self.project_id,
            'shares': {
                'in_use': 0,
                'limit': quotas['shares'],
                'reserved': 0,
            },
            'gigabytes': {
                'in_use': 0,
                'limit': quotas['gigabytes'],
                'reserved': 0,
            },
            'snapshots': {
                'in_use': 0,
                'limit': quotas['snapshots'],
                'reserved': 0,
            },
            'snapshot_gigabytes': {
                'in_use': 0,
                'limit': quotas['snapshot_gigabytes'],
                'reserved': 0,
            },
        }}

        for k, v in quotas.items():
            CONF.set_default('quota_' + k, v)

        result = self.controller.detail(req, self.project_id)

        self.assertEqual(expected, result)
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'show')
        quota_sets.db.share_type_get_by_name_or_id.assert_called_once_with(
            req.environ['manila.context'], 'fake_share_type_name_or_id')

    @ddt.data('2.39', '2.40')
    def test_show_share_type_quota(self, microversion):
        self.mock_object(
            quota_sets.db, 'share_type_get_by_name_or_id',
            mock.Mock(return_value={'id': 'fake_st_id'}))
        req = self._get_share_type_request_object(microversion)
        quotas = {
            "shares": 23,
            "snapshots": 34,
            "gigabytes": 45,
            "snapshot_gigabytes": 56,
        }
        expected = {
            'quota_set': {
                'id': self.project_id,
                'shares': quotas.get('shares', 50),
                'gigabytes': quotas.get('gigabytes', 1000),
                'snapshots': quotas.get('snapshots', 50),
                'snapshot_gigabytes': quotas.get('snapshot_gigabytes', 1000),
            }
        }
        for k, v in quotas.items():
            CONF.set_default('quota_' + k, v)

        result = self.controller.show(req, self.project_id)

        self.assertEqual(expected, result)
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'show')
        quota_sets.db.share_type_get_by_name_or_id.assert_called_once_with(
            req.environ['manila.context'], 'fake_share_type_name_or_id')

    @ddt.data('show', 'detail')
    def test_get_share_type_quota_with_old_microversion(self, method):
        req = self._get_share_type_request_object('2.38')
        self.assertRaises(
            webob.exc.HTTPBadRequest,
            getattr(self.controller, method),
            req, self.project_id)

    @ddt.data((None, None), (None, 'foo'), ('bar', None))
    @ddt.unpack
    def test__validate_user_id_and_share_type_args(self, user_id, st_id):
        result = self.controller._validate_user_id_and_share_type_args(
            user_id, st_id)

        self.assertIsNone(result)

    def test__validate_user_id_and_share_type_args_exception(self):
        self.assertRaises(
            webob.exc.HTTPBadRequest,
            self.controller._validate_user_id_and_share_type_args,
            'foo', 'bar')

    def test__get_share_type_id_found(self):
        self.mock_object(
            quota_sets.db, 'share_type_get_by_name_or_id',
            mock.Mock(return_value={'id': 'fake_st_id'}))
        ctxt = 'fake_context'
        share_type = 'fake_share_type_name_or_id'

        result = self.controller._get_share_type_id(ctxt, share_type)

        self.assertEqual('fake_st_id', result)

    def test__get_share_type_id_not_found(self):
        self.mock_object(
            quota_sets.db, 'share_type_get_by_name_or_id',
            mock.Mock(return_value=None))
        ctxt = 'fake_context'
        share_type = 'fake_share_type_name_or_id'

        self.assertRaises(
            webob.exc.HTTPNotFound,
            self.controller._get_share_type_id,
            ctxt, share_type)

    def test__get_share_type_id_is_not_provided(self):
        self.mock_object(
            quota_sets.db, 'share_type_get_by_name_or_id',
            mock.Mock(return_value={'id': 'fake_st_id'}))
        ctxt = 'fake_context'

        result = self.controller._get_share_type_id(ctxt, None)

        self.assertIsNone(result)

    @ddt.data(
        {},
        {"quota_set": {}},
        {"quota_set": {"foo": "bar"}},
        {"foo": "bar"},
    )
    def test__ensure_share_group_related_args_are_absent_success(self, body):
        result = self.controller._ensure_share_group_related_args_are_absent(
            body)

        self.assertIsNone(result)

    @ddt.data(
        {"share_groups": 5},
        {"share_group_snapshots": 6},
        {"quota_set": {"share_groups": 7}},
        {"quota_set": {"share_group_snapshots": 8}},
    )
    def test__ensure_share_group_related_args_are_absent_error(self, body):
        self.assertRaises(
            webob.exc.HTTPBadRequest,
            self.controller._ensure_share_group_related_args_are_absent, body)

    @ddt.data(_get_request(True, True), _get_request(True, False))
    def test__ensure_share_type_arg_is_absent(self, req):
        result = self.controller._ensure_share_type_arg_is_absent(req)

        self.assertIsNone(result)

    def test__ensure_share_type_arg_is_absent_exception(self):
        req = self._get_share_type_request_object('2.39')

        self.assertRaises(
            webob.exc.HTTPBadRequest,
            self.controller._ensure_share_type_arg_is_absent,
            req)

    @ddt.data(_get_request(True, True), _get_request(True, False))
    def test_quota_detail(self, request):
        request.api_version_request = api_version.APIVersionRequest('2.25')
        quotas = {
            "shares": 23,
            "snapshots": 34,
            "gigabytes": 45,
            "snapshot_gigabytes": 56,
            "share_networks": 67,
        }
        expected = {
            'quota_set': {
                'id': self.project_id,
                'shares': {'in_use': 0,
                           'limit': quotas['shares'],
                           'reserved': 0},
                'gigabytes': {'in_use': 0,
                              'limit': quotas['gigabytes'], 'reserved': 0},
                'snapshots': {'in_use': 0,
                              'limit': quotas['snapshots'], 'reserved': 0},
                'snapshot_gigabytes': {
                    'in_use': 0,
                    'limit': quotas['snapshot_gigabytes'],
                    'reserved': 0,
                },
                'share_networks': {
                    'in_use': 0,
                    'limit': quotas['share_networks'],
                    'reserved': 0
                },
            }
        }

        for k, v in quotas.items():
            CONF.set_default('quota_' + k, v)

        result = self.controller.detail(request, self.project_id)

        self.assertEqual(expected, result)
        self.mock_policy_check.assert_called_once_with(
            request.environ['manila.context'], self.resource_name, 'show')

    @ddt.data(_get_request(True, True), _get_request(True, False))
    def test_show_quota(self, request):
        quotas = {
            "shares": 23,
            "snapshots": 34,
            "gigabytes": 45,
            "snapshot_gigabytes": 56,
            "share_networks": 67,
            "share_groups": 53,
            "share_group_snapshots": 57,
        }
        expected = {
            'quota_set': {
                'id': self.project_id,
                'shares': quotas.get('shares', 50),
                'gigabytes': quotas.get('gigabytes', 1000),
                'snapshots': quotas.get('snapshots', 50),
                'snapshot_gigabytes': quotas.get('snapshot_gigabytes', 1000),
                'share_networks': quotas.get('share_networks', 10),
                'share_groups': quotas.get('share_groups', 50),
                'share_group_snapshots': quotas.get(
                    'share_group_snapshots', 50),
            }
        }
        for k, v in quotas.items():
            CONF.set_default('quota_' + k, v)

        result = self.controller.show(request, self.project_id)

        self.assertEqual(expected, result)
        self.mock_policy_check.assert_called_once_with(
            request.environ['manila.context'], self.resource_name, 'show')

    def test_show_quota_not_authorized(self):
        req = _get_request(True, False)
        self.mock_object(
            quota_sets.db,
            'authorize_project_context',
            mock.Mock(side_effect=exception.NotAuthorized))

        self.assertRaises(
            webob.exc.HTTPForbidden,
            self.controller.show,
            req, self.project_id)
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'show')

    @ddt.data(_get_request(True, True), _get_request(True, False))
    def test_update_quota(self, request):
        self.mock_object(
            quota_sets.db, 'share_type_get_by_name_or_id',
            mock.Mock(
                return_value={'id': 'fake_st_id', 'name': 'fake_st_name'}))
        CONF.set_default('quota_shares', 789)
        body = {'quota_set': {'tenant_id': self.project_id, 'shares': 788}}
        expected = {
            'quota_set': {
                'shares': body['quota_set']['shares'],
                'gigabytes': 1000,
                'snapshots': 50,
                'snapshot_gigabytes': 1000,
                'share_networks': 10,
                'share_groups': 50,
                'share_group_snapshots': 50,
            }
        }
        mock_policy_update_check_call = mock.call(
            request.environ['manila.context'], self.resource_name, 'update')
        mock_policy_show_check_call = mock.call(
            request.environ['manila.context'], self.resource_name, 'show')

        update_result = self.controller.update(
            request, self.project_id, body=body)

        self.assertEqual(expected, update_result)

        show_result = self.controller.show(request, self.project_id)

        expected['quota_set']['id'] = self.project_id
        self.assertEqual(expected, show_result)
        self.mock_policy_check.assert_has_calls([
            mock_policy_update_check_call, mock_policy_show_check_call])
        quota_sets.db.share_type_get_by_name_or_id.assert_not_called()

    @ddt.data('2.39', '2.40')
    def test_update_share_type_quota(self, microversion):
        self.mock_object(
            quota_sets.db, 'share_type_get_by_name_or_id',
            mock.Mock(
                return_value={'id': 'fake_st_id', 'name': 'fake_st_name'}))
        req = self._get_share_type_request_object(microversion)

        CONF.set_default('quota_shares', 789)
        body = {'quota_set': {'tenant_id': self.project_id, 'shares': 788}}
        expected = {
            'quota_set': {
                'shares': body['quota_set']['shares'],
                'gigabytes': 1000,
                'snapshots': 50,
                'snapshot_gigabytes': 1000,
            }
        }

        update_result = self.controller.update(req, self.project_id, body=body)

        self.assertEqual(expected, update_result)
        quota_sets.db.share_type_get_by_name_or_id.assert_called_once_with(
            req.environ['manila.context'],
            req.environ['QUERY_STRING'].split('=')[-1])
        quota_sets.db.share_type_get_by_name_or_id.reset_mock()

        show_result = self.controller.show(req, self.project_id)

        expected['quota_set']['id'] = self.project_id
        self.assertEqual(expected, show_result)
        self.mock_policy_check.assert_has_calls([
            mock.call(req.environ['manila.context'], self.resource_name, key)
            for key in ('update', 'show')
        ])
        quota_sets.db.share_type_get_by_name_or_id.assert_called_once_with(
            req.environ['manila.context'],
            req.environ['QUERY_STRING'].split('=')[-1])

    def test_update_share_type_quota_using_too_old_microversion(self):
        self.mock_object(
            quota_sets.db, 'share_type_get_by_name_or_id',
            mock.Mock(
                return_value={'id': 'fake_st_id', 'name': 'fake_st_name'}))
        req = self._get_share_type_request_object('2.38')
        body = {'quota_set': {'tenant_id': self.project_id, 'shares': 788}}

        self.assertRaises(
            webob.exc.HTTPBadRequest,
            self.controller.update,
            req, self.project_id, body=body)

        quota_sets.db.share_type_get_by_name_or_id.assert_not_called()

    def test_update_share_type_quota_for_share_networks(self):
        self.mock_object(
            quota_sets.db, 'share_type_get_by_name_or_id',
            mock.Mock(
                return_value={'id': 'fake_st_id', 'name': 'fake_st_name'}))
        req = self._get_share_type_request_object('2.39')
        body = {'quota_set': {
            'tenant_id': self.project_id, 'share_networks': 788,
        }}

        self.assertRaises(
            webob.exc.HTTPBadRequest,
            self.controller.update,
            req, self.project_id, body=body)

        quota_sets.db.share_type_get_by_name_or_id.assert_called_once_with(
            req.environ['manila.context'],
            req.environ['QUERY_STRING'].split('=')[-1])

    @ddt.data(-2, 'foo', {1: 2}, [1])
    def test_update_quota_with_invalid_value(self, value):
        req = _get_request(True, False)
        body = {'quota_set': {'tenant_id': self.project_id, 'shares': value}}

        self.assertRaises(
            webob.exc.HTTPBadRequest,
            self.controller.update,
            req, self.project_id, body=body)
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'update')

    def test_user_quota_can_not_be_bigger_than_tenant_quota(self):
        value = 777
        CONF.set_default('quota_shares', value)
        body = {
            'quota_set': {
                'tenant_id': self.project_id,
                'shares': value + 1,
            }
        }
        req = _get_request(True, True)

        self.assertRaises(
            webob.exc.HTTPBadRequest,
            self.controller.update,
            req, self.project_id, body=body)
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'update')

    def test_update_inexistent_quota(self):
        body = {
            'quota_set': {
                'tenant_id': self.project_id,
                'fake_quota': 13,
            }
        }
        req = _get_request(True, False)

        self.assertRaises(
            webob.exc.HTTPBadRequest,
            self.controller.update,
            req, self.project_id, body=body)
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'update')

    def test_update_quota_not_authorized(self):
        body = {'quota_set': {'tenant_id': self.project_id, 'shares': 13}}
        req = _get_request(False, False)

        self.assertRaises(
            webob.exc.HTTPForbidden,
            self.controller.update,
            req, self.project_id, body=body)
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'update')

    @ddt.data(
        ('os-quota-sets', '1.0', quota_sets.QuotaSetsControllerLegacy),
        ('os-quota-sets', '2.6', quota_sets.QuotaSetsControllerLegacy),
        ('quota-sets', '2.7', quota_sets.QuotaSetsController),
    )
    @ddt.unpack
    def test_update_all_quotas_with_force(self, url, version, controller):
        req = fakes.HTTPRequest.blank(
            '/fooproject/%s' % url, version=version, use_admin_context=True)
        quotas = (
            ('quota_shares', 13),
            ('quota_gigabytes', 14),
            ('quota_snapshots', 15),
            ('quota_snapshot_gigabytes', 16),
            ('quota_share_networks', 17),
        )
        for quota, value in quotas:
            CONF.set_default(quota, value)
        expected = {
            'quota_set': {
                'tenant_id': self.project_id,
                'shares': quotas[0][1],
                'gigabytes': quotas[1][1],
                'snapshots': quotas[2][1],
                'snapshot_gigabytes': quotas[3][1],
                'share_networks': quotas[4][1],
                'force': True,
            }
        }

        update_result = controller().update(
            req, self.project_id, body=expected)

        expected['quota_set'].pop('force')
        expected['quota_set'].pop('tenant_id')
        self.assertEqual(expected, update_result)

        show_result = controller().show(req, self.project_id)

        expected['quota_set']['id'] = self.project_id
        self.assertEqual(expected, show_result)
        self.mock_policy_check.assert_has_calls([
            mock.call(req.environ['manila.context'],
                      self.resource_name, action)
            for action in ('update', 'show')
        ])

    @ddt.data(
        ('os-quota-sets', '1.0', quota_sets.QuotaSetsControllerLegacy),
        ('os-quota-sets', '2.6', quota_sets.QuotaSetsControllerLegacy),
        ('quota-sets', '2.7', quota_sets.QuotaSetsController),
    )
    @ddt.unpack
    def test_delete_tenant_quota(self, url, version, controller):
        self.mock_object(quota_sets.QUOTAS, 'destroy_all_by_project_and_user')
        self.mock_object(quota_sets.QUOTAS, 'destroy_all_by_project')
        req = fakes.HTTPRequest.blank(
            '/fooproject/%s' % url, version=version, use_admin_context=True)

        result = controller().delete(req, self.project_id)

        self.assertTrue(
            utils.IsAMatcher(webob.response.Response) == result
        )
        self.assertTrue(hasattr(result, 'status_code'))
        self.assertEqual(202, result.status_code)
        self.assertFalse(
            quota_sets.QUOTAS.destroy_all_by_project_and_user.called)
        quota_sets.QUOTAS.destroy_all_by_project.assert_called_once_with(
            req.environ['manila.context'], self.project_id)
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'delete')

    def test_delete_user_quota(self):
        project_id = 'foo_project_id'
        self.mock_object(quota_sets.QUOTAS, 'destroy_all_by_project_and_user')
        self.mock_object(quota_sets.QUOTAS, 'destroy_all_by_project')
        req = _get_request(True, True)

        result = self.controller.delete(req, project_id)

        self.assertTrue(
            utils.IsAMatcher(webob.response.Response) == result
        )
        self.assertTrue(hasattr(result, 'status_code'))
        self.assertEqual(202, result.status_code)
        (quota_sets.QUOTAS.destroy_all_by_project_and_user.
            assert_called_once_with(
                req.environ['manila.context'],
                project_id,
                req.environ['manila.context'].user_id))
        self.assertFalse(quota_sets.QUOTAS.destroy_all_by_project.called)
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'delete')

    def test_delete_share_type_quota(self):
        req = self._get_share_type_request_object('2.39')
        self.mock_object(quota_sets.QUOTAS, 'destroy_all_by_project')
        self.mock_object(quota_sets.QUOTAS, 'destroy_all_by_project_and_user')
        mock_delete_st_quotas = self.mock_object(
            quota_sets.QUOTAS, 'destroy_all_by_project_and_share_type')
        self.mock_object(
            quota_sets.db, 'share_type_get_by_name_or_id',
            mock.Mock(
                return_value={'id': 'fake_st_id', 'name': 'fake_st_name'}))

        result = self.controller.delete(req, self.project_id)

        self.assertEqual(utils.IsAMatcher(webob.response.Response), result)
        self.assertTrue(hasattr(result, 'status_code'))
        self.assertEqual(202, result.status_code)
        mock_delete_st_quotas.assert_called_once_with(
            req.environ['manila.context'], self.project_id, 'fake_st_id')
        quota_sets.QUOTAS.destroy_all_by_project.assert_not_called()
        quota_sets.QUOTAS.destroy_all_by_project_and_user.assert_not_called()
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'delete')
        quota_sets.db.share_type_get_by_name_or_id.assert_called_once_with(
            req.environ['manila.context'],
            req.environ['QUERY_STRING'].split('=')[-1])

    def test_delete_share_type_quota_using_too_old_microversion(self):
        self.mock_object(
            quota_sets.db, 'share_type_get_by_name_or_id',
            mock.Mock(
                return_value={'id': 'fake_st_id', 'name': 'fake_st_name'}))
        req = self._get_share_type_request_object('2.38')

        self.assertRaises(
            webob.exc.HTTPBadRequest,
            self.controller.delete,
            req, self.project_id)

        quota_sets.db.share_type_get_by_name_or_id.assert_not_called()

    def test_delete_not_authorized(self):
        req = _get_request(False, False)
        self.assertRaises(
            webob.exc.HTTPForbidden,
            self.controller.delete,
            req, self.project_id)
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'delete')

    @ddt.data(
        ('os-quota-sets', '2.7', quota_sets.QuotaSetsControllerLegacy),
        ('quota-sets', '2.6', quota_sets.QuotaSetsController),
        ('quota-sets', '2.0', quota_sets.QuotaSetsController),
    )
    @ddt.unpack
    def test_api_not_found(self, url, version, controller):
        req = fakes.HTTPRequest.blank('/fooproject/%s' % url, version=version)
        for method_name in ('show', 'defaults', 'delete'):
            self.assertRaises(
                exception.VersionNotFoundForAPIMethod,
                getattr(controller(), method_name),
                req, self.project_id)

    @ddt.data(
        ('os-quota-sets', '2.7', quota_sets.QuotaSetsControllerLegacy),
        ('quota-sets', '2.6', quota_sets.QuotaSetsController),
        ('quota-sets', '2.0', quota_sets.QuotaSetsController),
    )
    @ddt.unpack
    def test_update_api_not_found(self, url, version, controller):
        req = fakes.HTTPRequest.blank('/fooproject/%s' % url, version=version)
        self.assertRaises(
            exception.VersionNotFoundForAPIMethod,
            controller().update,
            req, self.project_id)



# Copyright 2016 Mirantis inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Dummy share driver for testing Manila APIs and other interfaces.

This driver simulates support of:
- Both available driver modes: DHSS=True/False
- NFS and CIFS protocols
- IP access for NFS shares and USER access for CIFS shares
- CIFS shares in DHSS=True driver mode
- Creation and deletion of share snapshots
- Share replication (readable)
- Share migration
- Consistency groups
- Resize of a share (extend/shrink)

"""

import functools
import time

from oslo_config import cfg
from oslo_log import log
from oslo_utils import timeutils

from manila.common import constants
from manila import exception
from manila.i18n import _
from manila.share import configuration
from manila.share import driver
from manila.share.manager import share_manager_opts  # noqa
from manila.share import utils as share_utils

LOG = log.getLogger(__name__)


dummy_opts = [
    cfg.FloatOpt(
        "dummy_driver_default_driver_method_delay",
        help="Defines default time delay in seconds for each dummy driver "
             "method. To redefine some specific method delay use other "
             "'dummy_driver_driver_methods_delays' config opt. Optional.",
        default=2.0,
        min=0,
    ),
    cfg.DictOpt(
        "dummy_driver_driver_methods_delays",
        help="It is dictionary-like config option, that consists of "
             "driver method names as keys and integer/float values that are "
             "time delay in seconds. Optional.",
        default={
            "ensure_share": "1.05",
            "create_share": "3.98",
            "get_pool": "0.5",
            "do_setup": "0.05",

            "_get_pools_info": "0.1",
            "_update_share_stats": "0.3",

            "create_replica": "3.99",
            "delete_replica": "2.98",
            "promote_replica": "0.75",
            "update_replica_state": "0.85",
            "create_replicated_snapshot": "4.15",
            "delete_replicated_snapshot": "3.16",
            "update_replicated_snapshot": "1.17",

            "migration_start": 1.01,
            "migration_continue": 1.02,  # it will be called 2 times
            "migration_complete": 1.03,
            "migration_cancel": 1.04,
            "migration_get_progress": 1.05,
            "migration_check_compatibility": 0.05,
        },
    ),
]

CONF = cfg.CONF


def slow_me_down(f):

    @functools.wraps(f)
    def wrapped_func(self, *args, **kwargs):
        sleep_time = self.configuration.safe_get(
            "dummy_driver_driver_methods_delays").get(
                f.__name__,
                self.configuration.safe_get(
                    "dummy_driver_default_driver_method_delay")
            )
        time.sleep(float(sleep_time))
        return f(self, *args, **kwargs)

    return wrapped_func


def get_backend_configuration(backend_name):
    config_stanzas = CONF.list_all_sections()
    if backend_name not in config_stanzas:
        msg = _("Could not find backend stanza %(backend_name)s in "
                "configuration which is required for share replication and "
                "migration. Available stanzas are %(stanzas)s")
        params = {
            "stanzas": config_stanzas,
            "backend_name": backend_name,
        }
        raise exception.BadConfigurationException(reason=msg % params)

    config = configuration.Configuration(
        driver.share_opts, config_group=backend_name)
    config.append_config_values(dummy_opts)
    config.append_config_values(share_manager_opts)
    config.append_config_values(driver.ssh_opts)

    return config


class DummyDriver(driver.ShareDriver):
    """Dummy share driver that implements all share driver interfaces."""

    def __init__(self, *args, **kwargs):
        """Do initialization."""
        super(DummyDriver, self).__init__(
            [False, True], *args, config_opts=[dummy_opts], **kwargs)
        self._verify_configuration()
        self.private_storage = kwargs.get('private_storage')
        self.backend_name = self.configuration.safe_get(
            "share_backend_name") or "DummyDriver"
        self.migration_progress = {}

    def _verify_configuration(self):
        allowed_driver_methods = [m for m in dir(self) if m[0] != '_']
        allowed_driver_methods.extend([
            "_setup_server",
            "_teardown_server",
            "_get_pools_info",
            "_update_share_stats",
        ])
        disallowed_driver_methods = (
            "get_admin_network_allocations_number",
            "get_network_allocations_number",
            "get_share_server_pools",
        )
        for k, v in self.configuration.safe_get(
                "dummy_driver_driver_methods_delays").items():
            if k not in allowed_driver_methods:
                raise exception.BadConfigurationException(reason=(
                    "Dummy driver does not have '%s' method." % k
                ))
            elif k in disallowed_driver_methods:
                raise exception.BadConfigurationException(reason=(
                    "Method '%s' does not support delaying." % k
                ))
            try:
                float(v)
            except (TypeError, ValueError):
                raise exception.BadConfigurationException(reason=(
                    "Wrong value (%(v)s) for '%(k)s' dummy driver method time "
                    "delay is set in 'dummy_driver_driver_methods_delays' "
                    "config option." % {"k": k, "v": v}
                ))

    def _get_share_name(self, share):
        return "share_%(s_id)s_%(si_id)s" % {
            "s_id": share["share_id"].replace("-", "_"),
            "si_id": share["id"].replace("-", "_")}

    def _get_snapshot_name(self, snapshot):
        return "snapshot_%(s_id)s_%(si_id)s" % {
            "s_id": snapshot["snapshot_id"].replace("-", "_"),
            "si_id": snapshot["id"].replace("-", "_")}

    def _generate_export_locations(self, mountpoint, share_server=None):
        details = share_server["backend_details"] if share_server else {
            "primary_public_ip": "10.0.0.10",
            "secondary_public_ip": "10.0.0.20",
            "service_ip": "11.0.0.11",
        }
        return [
            {
                "path": "%(ip)s:%(mp)s" % {"ip": ip, "mp": mountpoint},
                "metadata": {
                    "preferred": preferred,
                },
                "is_admin_only": is_admin_only,
            } for ip, is_admin_only, preferred in (
                (details["primary_public_ip"], False, True),
                (details["secondary_public_ip"], False, False),
                (details["service_ip"], True, False))
        ]

    def _create_share(self, share, share_server=None):
        share_proto = share["share_proto"]
        if share_proto not in ("NFS", "CIFS"):
            msg = _("Unsupported share protocol provided - %s.") % share_proto
            raise exception.InvalidShareAccess(reason=msg)

        share_name = self._get_share_name(share)
        mountpoint = "/path/to/fake/share/%s" % share_name
        self.private_storage.update(
            share["id"], {
                "fake_provider_share_name": share_name,
                "fake_provider_location": mountpoint,
            }
        )
        return self._generate_export_locations(
            mountpoint, share_server=share_server)

    @slow_me_down
    def create_share(self, context, share, share_server=None):
        """Is called to create share."""
        return self._create_share(share, share_server=share_server)

    @slow_me_down
    def create_share_from_snapshot(self, context, share, snapshot,
                                   share_server=None):
        """Is called to create share from snapshot."""
        return self._create_share(share, share_server=share_server)

    def _create_snapshot(self, snapshot, share_server=None):
        snapshot_name = self._get_snapshot_name(snapshot)
        mountpoint = "/path/to/fake/snapshot/%s" % snapshot_name
        self.private_storage.update(
            snapshot["id"], {
                "fake_provider_snapshot_name": snapshot_name,
                "fake_provider_location": mountpoint,
            }
        )
        return {
            'fake_key1': 'fake_value1',
            'fake_key2': 'fake_value2',
            'fake_key3': 'fake_value3',
            "provider_location": mountpoint,
            "export_locations": self._generate_export_locations(
                mountpoint, share_server=share_server)
        }

    @slow_me_down
    def create_snapshot(self, context, snapshot, share_server=None):
        """Is called to create snapshot."""
        return self._create_snapshot(snapshot, share_server)

    @slow_me_down
    def delete_share(self, context, share, share_server=None):
        """Is called to remove share."""
        self.private_storage.delete(share["id"])

    @slow_me_down
    def delete_snapshot(self, context, snapshot, share_server=None):
        """Is called to remove snapshot."""
        LOG.debug('Deleting snapshot with following data: %s', snapshot)
        self.private_storage.delete(snapshot["id"])

    @slow_me_down
    def get_pool(self, share):
        """Return pool name where the share resides on."""
        pool_name = share_utils.extract_host(share["host"], level="pool")
        return pool_name

    @slow_me_down
    def ensure_share(self, context, share, share_server=None):
        """Invoked to ensure that share is exported."""

    @slow_me_down
    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        """Update access rules for given share."""
        for rule in add_rules + access_rules:
            share_proto = share["share_proto"].lower()
            access_type = rule["access_type"].lower()
            if not (
                    (share_proto == "nfs" and access_type == "ip") or
                    (share_proto == "cifs" and access_type == "user")):
                msg = _("Unsupported '%(access_type)s' access type provided "
                        "for '%(share_proto)s' share protocol.") % {
                    "access_type": access_type, "share_proto": share_proto}
                raise exception.InvalidShareAccess(reason=msg)

    @slow_me_down
    def snapshot_update_access(self, context, snapshot, access_rules,
                               add_rules, delete_rules, share_server=None):
        """Update access rules for given snapshot."""
        self.update_access(context, snapshot['share'], access_rules,
                           add_rules, delete_rules, share_server)

    @slow_me_down
    def do_setup(self, context):
        """Any initialization the share driver does while starting."""

    @slow_me_down
    def manage_existing(self, share, driver_options):
        """Brings an existing share under Manila management."""
        new_export = share['export_location']
        old_share_id = self._get_share_id_from_export(new_export)
        old_export = self.private_storage.get(
            old_share_id, key='export_location')
        if old_export.split(":/")[-1] == new_export.split(":/")[-1]:
            result = {"size": 1, "export_locations": self._create_share(share)}
            self.private_storage.delete(old_share_id)
            return result
        else:
            msg = ("Invalid export specified, existing share %s"
                   " could not be found" % old_share_id)
            raise exception.ShareBackendException(msg=msg)

    @slow_me_down
    def manage_existing_with_server(
            self, share, driver_options, share_server=None):
        return self.manage_existing(share, driver_options)

    def _get_share_id_from_export(self, export_location):
        values = export_location.split('share_')
        if len(values) > 1:
            return values[1][37:].replace("_", "-")
        else:
            return export_location

    @slow_me_down
    def unmanage(self, share):
        """Removes the specified share from Manila management."""
        self.private_storage.update(
            share['id'], {'export_location': share['export_location']})

    @slow_me_down
    def unmanage_with_server(self, share, share_server=None):
        self.unmanage(share)

    @slow_me_down
    def manage_existing_snapshot_with_server(self, snapshot, driver_options,
                                             share_server=None):
        return self.manage_existing_snapshot(snapshot, driver_options)

    @slow_me_down
    def manage_existing_snapshot(self, snapshot, driver_options):
        """Brings an existing snapshot under Manila management."""
        old_snap_id = self._get_snap_id_from_provider_location(
            snapshot['provider_location'])
        old_provider_location = self.private_storage.get(
            old_snap_id, key='provider_location')
        if old_provider_location == snapshot['provider_location']:
            self._create_snapshot(snapshot)
            self.private_storage.delete(old_snap_id)
            return {"size": 1,
                    "provider_location": snapshot["provider_location"]}
        else:
            msg = ("Invalid provider location specified, existing snapshot %s"
                   " could not be found" % old_snap_id)
            raise exception.ShareBackendException(msg=msg)

    def _get_snap_id_from_provider_location(self, provider_location):
        values = provider_location.split('snapshot_')
        if len(values) > 1:
            return values[1][37:].replace("_", "-")
        else:
            return provider_location

    @slow_me_down
    def unmanage_snapshot(self, snapshot):
        """Removes the specified snapshot from Manila management."""
        self.private_storage.update(
            snapshot['id'],
            {'provider_location': snapshot['provider_location']})

    @slow_me_down
    def unmanage_snapshot_with_server(self, snapshot, share_server=None):
        self.unmanage_snapshot(snapshot)

    @slow_me_down
    def revert_to_snapshot(self, context, snapshot, share_access_rules,
                           snapshot_access_rules, share_server=None):
        """Reverts a share (in place) to the specified snapshot."""

    @slow_me_down
    def extend_share(self, share, new_size, share_server=None):
        """Extends size of existing share."""

    @slow_me_down
    def shrink_share(self, share, new_size, share_server=None):
        """Shrinks size of existing share."""

    def get_network_allocations_number(self):
        """Returns number of network allocations for creating VIFs."""
        return 2

    def get_admin_network_allocations_number(self):
        return 1

    @slow_me_down
    def _setup_server(self, network_info, metadata=None):
        """Sets up and configures share server with given network parameters.

        Redefine it within share driver when it is going to handle share
        servers.
        """
        server_details = {
            "primary_public_ip": network_info[
                "network_allocations"][0]["ip_address"],
            "secondary_public_ip": network_info[
                "network_allocations"][1]["ip_address"],
            "service_ip": network_info[
                "admin_network_allocations"][0]["ip_address"],
            "username": "fake_username",
            "server_id": network_info['server_id']
        }
        return server_details

    @slow_me_down
    def _teardown_server(self, server_details, security_services=None):
        """Tears down share server."""

    @slow_me_down
    def _get_pools_info(self):
        pools = [{
            "pool_name": "fake_pool_for_%s" % self.backend_name,
            "total_capacity_gb": 1230.0,
            "free_capacity_gb": 1210.0,
            "reserved_percentage": self.configuration.reserved_share_percentage
        }]
        if self.configuration.replication_domain:
            pools[0]["replication_type"] = "readable"
        return pools

    @slow_me_down
    def _update_share_stats(self, data=None):
        """Retrieve stats info from share group."""
        data = {
            "share_backend_name": self.backend_name,
            "storage_protocol": "NFS_CIFS",
            "reserved_percentage":
                self.configuration.reserved_share_percentage,
            "snapshot_support": True,
            "create_share_from_snapshot_support": True,
            "revert_to_snapshot_support": True,
            "mount_snapshot_support": True,
            "driver_name": "Dummy",
            "pools": self._get_pools_info(),
            "share_group_stats": {
                "consistent_snapshot_support": "pool",
            }
        }
        if self.configuration.replication_domain:
            data["replication_type"] = "readable"
        super(DummyDriver, self)._update_share_stats(data)

    def get_share_server_pools(self, share_server):
        """Return list of pools related to a particular share server."""
        return []

    @slow_me_down
    def create_consistency_group(self, context, cg_dict, share_server=None):
        """Create a consistency group."""
        LOG.debug(
            "Successfully created dummy Consistency Group with ID: %s.",
            cg_dict["id"])

    @slow_me_down
    def delete_consistency_group(self, context, cg_dict, share_server=None):
        """Delete a consistency group."""
        LOG.debug(
            "Successfully deleted dummy consistency group with ID %s.",
            cg_dict["id"])

    @slow_me_down
    def create_cgsnapshot(self, context, snap_dict, share_server=None):
        """Create a consistency group snapshot."""
        LOG.debug("Successfully created CG snapshot %s.", snap_dict["id"])
        return None, None

    @slow_me_down
    def delete_cgsnapshot(self, context, snap_dict, share_server=None):
        """Delete a consistency group snapshot."""
        LOG.debug("Successfully deleted CG snapshot %s.", snap_dict["id"])
        return None, None

    @slow_me_down
    def create_consistency_group_from_cgsnapshot(
            self, context, cg_dict, cgsnapshot_dict, share_server=None):
        """Create a consistency group from a cgsnapshot."""
        LOG.debug(
            ("Successfully created dummy Consistency Group (%(cg_id)s) "
             "from CG snapshot (%(cg_snap_id)s)."),
            {"cg_id": cg_dict["id"], "cg_snap_id": cgsnapshot_dict["id"]})
        return None, []

    @slow_me_down
    def create_replica(self, context, replica_list, new_replica,
                       access_rules, replica_snapshots, share_server=None):
        """Replicate the active replica to a new replica on this backend."""
        replica_name = self._get_share_name(new_replica)
        mountpoint = "/path/to/fake/share/%s" % replica_name
        self.private_storage.update(
            new_replica["id"], {
                "fake_provider_replica_name": replica_name,
                "fake_provider_location": mountpoint,
            }
        )
        return {
            "export_locations": self._generate_export_locations(
                mountpoint, share_server=share_server),
            "replica_state": constants.REPLICA_STATE_IN_SYNC,
            "access_rules_status": constants.STATUS_ACTIVE,
        }

    @slow_me_down
    def delete_replica(self, context, replica_list, replica_snapshots,
                       replica, share_server=None):
        """Delete a replica."""
        self.private_storage.delete(replica["id"])

    @slow_me_down
    def promote_replica(self, context, replica_list, replica, access_rules,
                        share_server=None):
        """Promote a replica to 'active' replica state."""
        return_replica_list = []
        for r in replica_list:
            if r["id"] == replica["id"]:
                replica_state = constants.REPLICA_STATE_ACTIVE
            else:
                replica_state = constants.REPLICA_STATE_IN_SYNC
            return_replica_list.append(
                {"id": r["id"], "replica_state": replica_state})
        return return_replica_list

    @slow_me_down
    def update_replica_state(self, context, replica_list, replica,
                             access_rules, replica_snapshots,
                             share_server=None):
        """Update the replica_state of a replica."""
        return constants.REPLICA_STATE_IN_SYNC

    @slow_me_down
    def create_replicated_snapshot(self, context, replica_list,
                                   replica_snapshots, share_server=None):
        """Create a snapshot on active instance and update across the replicas.

        """
        return_replica_snapshots = []
        for r in replica_snapshots:
            return_replica_snapshots.append(
                {"id": r["id"], "status": constants.STATUS_AVAILABLE})
        return return_replica_snapshots

    @slow_me_down
    def revert_to_replicated_snapshot(self, context, active_replica,
                                      replica_list, active_replica_snapshot,
                                      replica_snapshots, share_access_rules,
                                      snapshot_access_rules,
                                      share_server=None):
        """Reverts a replicated share (in place) to the specified snapshot."""

    @slow_me_down
    def delete_replicated_snapshot(self, context, replica_list,
                                   replica_snapshots, share_server=None):
        """Delete a snapshot by deleting its instances across the replicas."""
        return_replica_snapshots = []
        for r in replica_snapshots:
            return_replica_snapshots.append(
                {"id": r["id"], "status": constants.STATUS_DELETED})
        return return_replica_snapshots

    @slow_me_down
    def update_replicated_snapshot(self, context, replica_list,
                                   share_replica, replica_snapshots,
                                   replica_snapshot, share_server=None):
        """Update the status of a snapshot instance that lives on a replica."""
        return {
            "id": replica_snapshot["id"], "status": constants.STATUS_AVAILABLE}

    @slow_me_down
    def migration_check_compatibility(
            self, context, source_share, destination_share,
            share_server=None, destination_share_server=None):
        """Is called to test compatibility with destination backend."""
        backend_name = share_utils.extract_host(
            destination_share['host'], level='backend_name')
        config = get_backend_configuration(backend_name)
        compatible = 'Dummy' in config.share_driver
        return {
            'compatible': compatible,
            'writable': compatible,
            'preserve_metadata': compatible,
            'nondisruptive': False,
            'preserve_snapshots': compatible,
        }

    @slow_me_down
    def migration_start(
            self, context, source_share, destination_share, source_snapshots,
            snapshot_mappings, share_server=None,
            destination_share_server=None):
        """Is called to perform 1st phase of driver migration of a given share.

        """
        LOG.debug(
            "Migration of dummy share with ID '%s' has been started.",
            source_share["id"])
        self.migration_progress[source_share['share_id']] = 0

    @slow_me_down
    def migration_continue(
            self, context, source_share, destination_share, source_snapshots,
            snapshot_mappings, share_server=None,
            destination_share_server=None):

        if source_share["id"] not in self.migration_progress:
            self.migration_progress[source_share["id"]] = 0

        self.migration_progress[source_share["id"]] += 50

        LOG.debug(
            "Migration of dummy share with ID '%s' is continuing, %s.",
            source_share["id"],
            self.migration_progress[source_share["id"]])

        return self.migration_progress[source_share["id"]] == 100

    @slow_me_down
    def migration_complete(
            self, context, source_share, destination_share, source_snapshots,
            snapshot_mappings, share_server=None,
            destination_share_server=None):
        """Is called to perform 2nd phase of driver migration of a given share.

        """
        snapshot_updates = {}
        for src_snap_ins, dest_snap_ins in snapshot_mappings.items():
            snapshot_updates[dest_snap_ins['id']] = self._create_snapshot(
                dest_snap_ins)
        return {
            'snapshot_updates': snapshot_updates,
            'export_locations': self._do_migration(
                source_share, destination_share, share_server)
        }

    def _do_migration(self, source_share_ref, dest_share_ref, share_server):
        share_name = self._get_share_name(dest_share_ref)
        mountpoint = "/path/to/fake/share/%s" % share_name
        self.private_storage.delete(source_share_ref["id"])
        self.private_storage.update(
            dest_share_ref["id"], {
                "fake_provider_share_name": share_name,
                "fake_provider_location": mountpoint,
            }
        )
        LOG.debug(
            "Migration of dummy share with ID '%s' has been completed.",
            source_share_ref["id"])
        self.migration_progress.pop(source_share_ref["id"], None)

        return self._generate_export_locations(
            mountpoint, share_server=share_server)

    @slow_me_down
    def migration_cancel(
            self, context, source_share, destination_share, source_snapshots,
            snapshot_mappings, share_server=None,
            destination_share_server=None):
        """Is called to cancel driver migration."""
        LOG.debug(
            "Migration of dummy share with ID '%s' has been canceled.",
            source_share["id"])
        self.migration_progress.pop(source_share["id"], None)

    @slow_me_down
    def migration_get_progress(
            self, context, source_share, destination_share, source_snapshots,
            snapshot_mappings, share_server=None,
            destination_share_server=None):
        """Is called to get migration progress."""
        # Simulate migration progress.
        if source_share["id"] not in self.migration_progress:
            self.migration_progress[source_share["id"]] = 0
        total_progress = self.migration_progress[source_share["id"]]
        LOG.debug("Progress of current dummy share migration "
                  "with ID '%(id)s' is %(progress)s.", {
                      "id": source_share["id"],
                      "progress": total_progress
                  })
        return {"total_progress": total_progress}

    def update_share_usage_size(self, context, shares):
        share_updates = []
        gathered_at = timeutils.utcnow()
        for s in shares:
            share_updates.append({'id': s['id'],
                                  'used_size':  1,
                                  'gathered_at': gathered_at})
        return share_updates

    @slow_me_down
    def get_share_server_network_info(
            self, context, share_server, identifier, driver_options):
        try:
            server_details = self.private_storage.get(identifier)
        except Exception:
            msg = ("Unable to find share server %s in "
                   "private storage." % identifier)
            raise exception.ShareBackendException(msg=msg)

        return [server_details['primary_public_ip'],
                server_details['secondary_public_ip'],
                server_details['service_ip']]

    @slow_me_down
    def manage_server(self, context, share_server, identifier, driver_options):
        server_details = self.private_storage.get(identifier)
        self.private_storage.delete(identifier)
        return identifier, server_details

    def unmanage_server(self, server_details, security_services=None):
        server_details = server_details or {}
        if not server_details or 'server_id' not in server_details:
            # This share server doesn't have any network details. Since it's
            # just being cleaned up, we'll log a warning and return without
            # errors.
            LOG.warning("Share server does not have network information. "
                        "It is being unmanaged, but cannot be re-managed "
                        "without first creating network allocations in this "
                        "driver's private storage.")
            return
        self.private_storage.update(server_details['server_id'],
                                    server_details)



# Copyright 2015 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import ddt
import mock

from manila import context
from manila.share import hook
from manila import test


class FakeHookImplementation(hook.HookBase):
    def _execute_pre_hook(self, context, func_name, *args, **kwargs):
        """Fake implementation of a pre hook action."""

    def _execute_post_hook(self, context, func_name, pre_hook_data,
                           driver_action_results, *args, **kwargs):
        """Fake implementation of a post hook action."""

    def _execute_periodic_hook(self, context, periodic_hook_data,
                               *args, **kwargs):
        """Fake implementation of a periodic hook action."""


@ddt.ddt
class HookBaseTestCase(test.TestCase):

    def setUp(self):
        super(HookBaseTestCase, self).setUp()
        self.context = context.get_admin_context()
        self.default_config = {
            "enable_pre_hooks": True,
            "enable_post_hooks": True,
            "enable_periodic_hooks": True,
            "suppress_pre_hooks_errors": True,
            "suppress_post_hooks_errors": True,
        }
        for k, v in self.default_config.items():
            hook.CONF.set_default(k, v)

    def _fake_safe_get(self, key):
        return self.default_config.get(key)

    def _get_hook_instance(self, set_configuration=True, host="fake_host"):
        if set_configuration:
            configuration = mock.Mock()
            configuration.safe_get.side_effect = self._fake_safe_get
        else:
            configuration = None
        instance = FakeHookImplementation(
            configuration=configuration, host=host)
        return instance

    def test_instantiate_hook_fail(self):
        self.assertRaises(TypeError, hook.HookBase)

    @ddt.data(True, False)
    def test_instantiate_hook_successfully_and_set_configuration(
            self, set_configuration):
        instance = self._get_hook_instance(set_configuration)

        self.assertTrue(hasattr(instance, 'host'))
        self.assertEqual("fake_host", instance.host)
        self.assertTrue(hasattr(instance, 'configuration'))
        if not set_configuration:
            self.assertIsNone(instance.configuration)
        for attr_name in ("pre_hooks_enabled",
                          "post_hooks_enabled",
                          "periodic_hooks_enabled",
                          "suppress_pre_hooks_errors",
                          "suppress_post_hooks_errors"):
            self.assertTrue(hasattr(instance, attr_name))
        if set_configuration:
            instance.configuration.append_config_values.assert_has_calls([
                mock.call(hook.hook_options)])
            conf_func = self._fake_safe_get
        else:
            conf_func = self.default_config.get
        self.assertEqual(
            conf_func("enable_pre_hooks"), instance.pre_hooks_enabled)
        self.assertEqual(
            conf_func("enable_post_hooks"), instance.post_hooks_enabled)
        self.assertEqual(
            conf_func("enable_periodic_hooks"),
            instance.periodic_hooks_enabled)
        self.assertEqual(
            conf_func("suppress_pre_hooks_errors"),
            instance.suppress_pre_hooks_errors)
        self.assertEqual(
            conf_func("suppress_post_hooks_errors"),
            instance.suppress_post_hooks_errors)

    def test_execute_pre_hook_disabled(self):
        instance = self._get_hook_instance()
        instance.pre_hooks_enabled = False
        self.mock_object(
            instance, "_execute_pre_hook",
            mock.Mock(side_effect=Exception("I should not be raised.")))

        result = instance.execute_pre_hook(
            self.context, "fake_func_name", "some_arg", some_kwarg="foo")

        self.assertIsNone(result)

    @ddt.data(True, False)
    def test_execute_pre_hook_success(self, provide_context):
        instance = self._get_hook_instance()
        instance.pre_hooks_enabled = True
        instance.suppress_pre_hooks_errors = True
        expected = "fake_expected_result"
        some_arg = "some_arg"
        func_name = "fake_func_name"
        self.mock_object(hook.LOG, 'error')
        self.mock_object(
            instance, "_execute_pre_hook", mock.Mock(return_value=expected))
        mock_ctxt = self.mock_object(context, 'get_admin_context')
        ctxt = self.context if provide_context else mock_ctxt

        result = instance.execute_pre_hook(
            ctxt, func_name, some_arg, some_kwarg="foo")

        self.assertEqual(expected, result)
        instance._execute_pre_hook.assert_called_once_with(
            some_arg,
            context=self.context if provide_context else mock_ctxt,
            func_name=func_name,
            some_kwarg="foo")
        self.assertFalse(hook.LOG.error.called)

    def test_execute_pre_hook_exception_with_suppression(self):
        instance = self._get_hook_instance()
        instance.pre_hooks_enabled = True
        instance.suppress_pre_hooks_errors = True
        some_arg = "some_arg"
        func_name = "fake_func_name"
        FakeException = type("FakeException", (Exception, ), {})
        self.mock_object(hook.LOG, 'warning')
        self.mock_object(
            instance, "_execute_pre_hook", mock.Mock(side_effect=(
                FakeException("Some exception that should be suppressed."))))

        result = instance.execute_pre_hook(
            self.context, func_name, some_arg, some_kwarg="foo")

        self.assertIsInstance(result, FakeException)
        instance._execute_pre_hook.assert_called_once_with(
            some_arg,
            context=self.context,
            func_name=func_name,
            some_kwarg="foo")
        self.assertTrue(hook.LOG.warning.called)

    def test_execute_pre_hook_exception_without_suppression(self):
        instance = self._get_hook_instance()
        instance.pre_hooks_enabled = True
        instance.suppress_pre_hooks_errors = False
        some_arg = "some_arg"
        func_name = "fake_func_name"
        FakeException = type("FakeException", (Exception, ), {})
        self.mock_object(hook.LOG, 'warning')
        self.mock_object(
            instance, "_execute_pre_hook", mock.Mock(side_effect=(
                FakeException(
                    "Some exception that should NOT be suppressed."))))

        self.assertRaises(
            FakeException,
            instance.execute_pre_hook,
            self.context, func_name, some_arg, some_kwarg="foo")

        instance._execute_pre_hook.assert_called_once_with(
            some_arg,
            context=self.context,
            func_name=func_name,
            some_kwarg="foo")
        self.assertFalse(hook.LOG.warning.called)

    def test_execute_post_hook_disabled(self):
        instance = self._get_hook_instance()
        instance.post_hooks_enabled = False
        self.mock_object(
            instance, "_execute_post_hook",
            mock.Mock(side_effect=Exception("I should not be raised.")))

        result = instance.execute_post_hook(
            self.context, "fake_func_name", "some_pre_hook_data",
            "some_driver_action_results", "some_arg", some_kwarg="foo")

        self.assertIsNone(result)

    @ddt.data(True, False)
    def test_execute_post_hook_success(self, provide_context):
        instance = self._get_hook_instance()
        instance.post_hooks_enabled = True
        instance.suppress_post_hooks_errors = True
        expected = "fake_expected_result"
        some_arg = "some_arg"
        func_name = "fake_func_name"
        pre_hook_data = "some_pre_hook_data"
        driver_action_results = "some_driver_action_results"
        self.mock_object(hook.LOG, 'warning')
        self.mock_object(
            instance, "_execute_post_hook", mock.Mock(return_value=expected))
        mock_ctxt = self.mock_object(context, 'get_admin_context')
        ctxt = self.context if provide_context else mock_ctxt

        result = instance.execute_post_hook(
            ctxt, func_name, pre_hook_data, driver_action_results,
            some_arg, some_kwarg="foo")

        self.assertEqual(expected, result)
        instance._execute_post_hook.assert_called_once_with(
            some_arg,
            context=self.context if provide_context else mock_ctxt,
            func_name=func_name,
            pre_hook_data=pre_hook_data,
            driver_action_results=driver_action_results,
            some_kwarg="foo")
        self.assertFalse(hook.LOG.warning.called)

    def test_execute_post_hook_exception_with_suppression(self):
        instance = self._get_hook_instance()
        instance.post_hooks_enabled = True
        instance.suppress_post_hooks_errors = True
        some_arg = "some_arg"
        func_name = "fake_func_name"
        pre_hook_data = "some_pre_hook_data"
        driver_action_results = "some_driver_action_results"
        FakeException = type("FakeException", (Exception, ), {})
        self.mock_object(hook.LOG, 'warning')
        self.mock_object(
            instance, "_execute_post_hook", mock.Mock(side_effect=(
                FakeException("Some exception that should be suppressed."))))

        result = instance.execute_post_hook(
            self.context, func_name, pre_hook_data, driver_action_results,
            some_arg, some_kwarg="foo")

        self.assertIsInstance(result, FakeException)
        instance._execute_post_hook.assert_called_once_with(
            some_arg,
            context=self.context,
            func_name=func_name,
            pre_hook_data=pre_hook_data,
            driver_action_results=driver_action_results,
            some_kwarg="foo")
        self.assertTrue(hook.LOG.warning.called)

    def test_execute_post_hook_exception_without_suppression(self):
        instance = self._get_hook_instance()
        instance.post_hooks_enabled = True
        instance.suppress_post_hooks_errors = False
        some_arg = "some_arg"
        func_name = "fake_func_name"
        pre_hook_data = "some_pre_hook_data"
        driver_action_results = "some_driver_action_results"
        FakeException = type("FakeException", (Exception, ), {})
        self.mock_object(hook.LOG, 'error')
        self.mock_object(
            instance, "_execute_post_hook", mock.Mock(side_effect=(
                FakeException(
                    "Some exception that should NOT be suppressed."))))

        self.assertRaises(
            FakeException,
            instance.execute_post_hook,
            self.context, func_name, pre_hook_data, driver_action_results,
            some_arg, some_kwarg="foo")

        instance._execute_post_hook.assert_called_once_with(
            some_arg,
            context=self.context,
            func_name=func_name,
            pre_hook_data=pre_hook_data,
            driver_action_results=driver_action_results,
            some_kwarg="foo")
        self.assertFalse(hook.LOG.error.called)

    def test_execute_periodic_hook_disabled(self):
        instance = self._get_hook_instance()
        instance.periodic_hooks_enabled = False
        self.mock_object(instance, "_execute_periodic_hook")

        instance.execute_periodic_hook(
            self.context, "fake_periodic_hook_data",
            "some_arg", some_kwarg="foo")

        self.assertFalse(instance._execute_periodic_hook.called)

    @ddt.data(True, False)
    def test_execute_periodic_hook_enabled(self, provide_context):
        instance = self._get_hook_instance()
        instance.periodic_hooks_enabled = True
        expected = "some_expected_result"
        self.mock_object(
            instance,
            "_execute_periodic_hook",
            mock.Mock(return_value=expected))
        mock_ctxt = self.mock_object(context, 'get_admin_context')
        ctxt = self.context if provide_context else mock_ctxt

        result = instance.execute_periodic_hook(
            ctxt, "fake_periodic_hook_data",
            "some_arg", some_kwarg="foo")

        instance._execute_periodic_hook.assert_called_once_with(
            ctxt, "fake_periodic_hook_data",
            "some_arg", some_kwarg="foo")
        self.assertEqual(expected, result)



# Copyright (c) 2011 X.commerce, a business unit of eBay Inc.
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Defines interface for DB access.

The underlying driver is loaded as a :class:`LazyPluggable`.

Functions in this module are imported into the manila.db namespace. Call these
functions from manila.db namespace, not the manila.db.api namespace.

All functions in this module return objects that implement a dictionary-like
interface. Currently, many of these objects are sqlalchemy objects that
implement a dictionary interface. However, a future goal is to have all of
these objects be simple dictionaries.


**Related Flags**

:backend:  string to lookup in the list of LazyPluggable backends.
           `sqlalchemy` is the only supported backend right now.

:connection:  string specifying the sqlalchemy connection to use, like:
              `sqlite:///var/lib/manila/manila.sqlite`.

:enable_new_services:  when adding a new service to the database, is it in the
                       pool of available hardware (Default: True)

"""
from oslo_config import cfg
from oslo_db import api as db_api

db_opts = [
    cfg.StrOpt('db_backend',
               default='sqlalchemy',
               help='The backend to use for database.'),
    cfg.BoolOpt('enable_new_services',
                default=True,
                help='Services to be added to the available pool on create.'),
    cfg.StrOpt('share_name_template',
               default='share-%s',
               help='Template string to be used to generate share names.'),
    cfg.StrOpt('share_snapshot_name_template',
               default='share-snapshot-%s',
               help='Template string to be used to generate share snapshot '
                    'names.'),
]

CONF = cfg.CONF
CONF.register_opts(db_opts)

_BACKEND_MAPPING = {'sqlalchemy': 'manila.db.sqlalchemy.api'}
IMPL = db_api.DBAPI.from_config(cfg.CONF, backend_mapping=_BACKEND_MAPPING,
                                lazy=True)


def authorize_project_context(context, project_id):
    """Ensures a request has permission to access the given project."""
    return IMPL.authorize_project_context(context, project_id)


def authorize_quota_class_context(context, class_name):
    """Ensures a request has permission to access the given quota class."""
    return IMPL.authorize_quota_class_context(context, class_name)


###################


def service_destroy(context, service_id):
    """Destroy the service or raise if it does not exist."""
    return IMPL.service_destroy(context, service_id)


def service_get(context, service_id):
    """Get a service or raise if it does not exist."""
    return IMPL.service_get(context, service_id)


def service_get_by_host_and_topic(context, host, topic):
    """Get a service by host it's on and topic it listens to."""
    return IMPL.service_get_by_host_and_topic(context, host, topic)


def service_get_all(context, disabled=None):
    """Get all services."""
    return IMPL.service_get_all(context, disabled)


def service_get_all_by_topic(context, topic):
    """Get all services for a given topic."""
    return IMPL.service_get_all_by_topic(context, topic)


def service_get_all_share_sorted(context):
    """Get all share services sorted by share count.

    :returns: a list of (Service, share_count) tuples.

    """
    return IMPL.service_get_all_share_sorted(context)


def service_get_by_args(context, host, binary):
    """Get the state of an service by node name and binary."""
    return IMPL.service_get_by_args(context, host, binary)


def service_create(context, values):
    """Create a service from the values dictionary."""
    return IMPL.service_create(context, values)


def service_update(context, service_id, values):
    """Set the given properties on an service and update it.

    Raises NotFound if service does not exist.

    """
    return IMPL.service_update(context, service_id, values)


####################


def quota_create(context, project_id, resource, limit, user_id=None,
                 share_type_id=None):
    """Create a quota for the given project and resource."""
    return IMPL.quota_create(context, project_id, resource, limit,
                             user_id=user_id, share_type_id=share_type_id)


def quota_get_all_by_project_and_user(context, project_id, user_id):
    """Retrieve all quotas associated with a given project and user."""
    return IMPL.quota_get_all_by_project_and_user(context, project_id, user_id)


def quota_get_all_by_project_and_share_type(context, project_id,
                                            share_type_id):
    """Retrieve all quotas associated with a given project and user."""
    return IMPL.quota_get_all_by_project_and_share_type(
        context, project_id, share_type_id)


def quota_get_all_by_project(context, project_id):
    """Retrieve all quotas associated with a given project."""
    return IMPL.quota_get_all_by_project(context, project_id)


def quota_get_all(context, project_id):
    """Retrieve all user quotas associated with a given project."""
    return IMPL.quota_get_all(context, project_id)


def quota_update(context, project_id, resource, limit, user_id=None,
                 share_type_id=None):
    """Update a quota or raise if it does not exist."""
    return IMPL.quota_update(context, project_id, resource, limit,
                             user_id=user_id, share_type_id=share_type_id)


###################


def quota_class_create(context, class_name, resource, limit):
    """Create a quota class for the given name and resource."""
    return IMPL.quota_class_create(context, class_name, resource, limit)


def quota_class_get(context, class_name, resource):
    """Retrieve a quota class or raise if it does not exist."""
    return IMPL.quota_class_get(context, class_name, resource)


def quota_class_get_default(context):
    """Retrieve all default quotas."""
    return IMPL.quota_class_get_default(context)


def quota_class_get_all_by_name(context, class_name):
    """Retrieve all quotas associated with a given quota class."""
    return IMPL.quota_class_get_all_by_name(context, class_name)


def quota_class_update(context, class_name, resource, limit):
    """Update a quota class or raise if it does not exist."""
    return IMPL.quota_class_update(context, class_name, resource, limit)


###################


def quota_usage_get(context, project_id, resource, user_id=None,
                    share_type_id=None):
    """Retrieve a quota usage or raise if it does not exist."""
    return IMPL.quota_usage_get(
        context, project_id, resource, user_id=user_id,
        share_type_id=share_type_id)


def quota_usage_get_all_by_project_and_user(context, project_id, user_id):
    """Retrieve all usage associated with a given resource."""
    return IMPL.quota_usage_get_all_by_project_and_user(context,
                                                        project_id, user_id)


def quota_usage_get_all_by_project_and_share_type(context, project_id,
                                                  share_type_id):
    """Retrieve all usage associated with a given resource."""
    return IMPL.quota_usage_get_all_by_project_and_share_type(
        context, project_id, share_type_id)


def quota_usage_get_all_by_project(context, project_id):
    """Retrieve all usage associated with a given resource."""
    return IMPL.quota_usage_get_all_by_project(context, project_id)


def quota_usage_create(context, project_id, user_id, resource, in_use,
                       reserved=0, until_refresh=None, share_type_id=None):
    """Create a quota usage."""
    return IMPL.quota_usage_create(
        context, project_id, user_id, resource, in_use, reserved,
        until_refresh, share_type_id=share_type_id)


def quota_usage_update(context, project_id, user_id, resource,
                       share_type_id=None, **kwargs):
    """Update a quota usage or raise if it does not exist."""
    return IMPL.quota_usage_update(
        context, project_id, user_id, resource, share_type_id=share_type_id,
        **kwargs)


###################


def quota_reserve(context, resources, quotas, user_quotas, share_type_quotas,
                  deltas, expire, until_refresh, max_age,
                  project_id=None, user_id=None, share_type_id=None):
    """Check quotas and create appropriate reservations."""
    return IMPL.quota_reserve(
        context, resources, quotas, user_quotas, share_type_quotas, deltas,
        expire, until_refresh, max_age, project_id=project_id, user_id=user_id,
        share_type_id=share_type_id)


def reservation_commit(context, reservations, project_id=None, user_id=None,
                       share_type_id=None):
    """Commit quota reservations."""
    return IMPL.reservation_commit(
        context, reservations, project_id=project_id, user_id=user_id,
        share_type_id=share_type_id)


def reservation_rollback(context, reservations, project_id=None, user_id=None,
                         share_type_id=None):
    """Roll back quota reservations."""
    return IMPL.reservation_rollback(
        context, reservations, project_id=project_id, user_id=user_id,
        share_type_id=share_type_id)


def quota_destroy_all_by_project_and_user(context, project_id, user_id):
    """Destroy all quotas associated with a given project and user."""
    return IMPL.quota_destroy_all_by_project_and_user(context,
                                                      project_id, user_id)


def quota_destroy_all_by_share_type(context, share_type_id, project_id=None):
    """Destroy all quotas associated with a given share type and project."""
    return IMPL.quota_destroy_all_by_share_type(
        context, share_type_id, project_id=project_id)


def quota_destroy_all_by_project(context, project_id):
    """Destroy all quotas associated with a given project."""
    return IMPL.quota_destroy_all_by_project(context, project_id)


def reservation_expire(context):
    """Roll back any expired reservations."""
    return IMPL.reservation_expire(context)


###################


def share_instance_get(context, instance_id, with_share_data=False):
    """Get share instance by id."""
    return IMPL.share_instance_get(context, instance_id,
                                   with_share_data=with_share_data)


def share_instance_create(context, share_id, values):
    """Create new share instance."""
    return IMPL.share_instance_create(context, share_id, values)


def share_instance_delete(context, instance_id, session=None,
                          need_to_update_usages=False):
    """Delete share instance."""
    return IMPL.share_instance_delete(
        context, instance_id, session=session,
        need_to_update_usages=need_to_update_usages)


def share_instance_update(context, instance_id, values, with_share_data=False):
    """Update share instance fields."""
    return IMPL.share_instance_update(context, instance_id, values,
                                      with_share_data=with_share_data)


def share_instances_host_update(context, current_host, new_host):
    """Update the host attr of all share instances that are on current_host."""
    return IMPL.share_instances_host_update(context, current_host, new_host)


def share_instances_get_all(context, filters=None):
    """Returns all share instances."""
    return IMPL.share_instances_get_all(context, filters=filters)


def share_instances_get_all_by_share_server(context, share_server_id):
    """Returns all share instances with given share_server_id."""
    return IMPL.share_instances_get_all_by_share_server(context,
                                                        share_server_id)


def share_instances_get_all_by_host(context, host, with_share_data=False):
    """Returns all share instances with given host."""
    return IMPL.share_instances_get_all_by_host(
        context, host, with_share_data=with_share_data)


def share_instances_get_all_by_share_network(context, share_network_id):
    """Returns list of shares that belong to given share network."""
    return IMPL.share_instances_get_all_by_share_network(context,
                                                         share_network_id)


def share_instances_get_all_by_share(context, share_id):
    """Returns list of shares that belong to given share."""
    return IMPL.share_instances_get_all_by_share(context, share_id)


def share_instances_get_all_by_share_group_id(context, share_group_id):
    """Returns list of share instances that belong to given share group."""
    return IMPL.share_instances_get_all_by_share_group_id(
        context, share_group_id)

###################


def share_create(context, share_values, create_share_instance=True):
    """Create new share."""
    return IMPL.share_create(context, share_values,
                             create_share_instance=create_share_instance)


def share_update(context, share_id, values):
    """Update share fields."""
    return IMPL.share_update(context, share_id, values)


def share_get(context, share_id):
    """Get share by id."""
    return IMPL.share_get(context, share_id)


def share_get_all(context, filters=None, sort_key=None, sort_dir=None):
    """Get all shares."""
    return IMPL.share_get_all(
        context, filters=filters, sort_key=sort_key, sort_dir=sort_dir,
    )


def share_get_all_by_project(context, project_id, filters=None,
                             is_public=False, sort_key=None, sort_dir=None):
    """Returns all shares with given project ID."""
    return IMPL.share_get_all_by_project(
        context, project_id, filters=filters, is_public=is_public,
        sort_key=sort_key, sort_dir=sort_dir,
    )


def share_get_all_by_share_group_id(context, share_group_id,
                                    filters=None, sort_key=None,
                                    sort_dir=None):
    """Returns all shares with given project ID and share group id."""
    return IMPL.share_get_all_by_share_group_id(
        context, share_group_id, filters=filters,
        sort_key=sort_key, sort_dir=sort_dir)


def share_get_all_by_share_server(context, share_server_id, filters=None,
                                  sort_key=None, sort_dir=None):
    """Returns all shares with given share server ID."""
    return IMPL.share_get_all_by_share_server(
        context, share_server_id, filters=filters, sort_key=sort_key,
        sort_dir=sort_dir,
    )


def share_delete(context, share_id):
    """Delete share."""
    return IMPL.share_delete(context, share_id)


###################


def share_access_create(context, values):
    """Allow access to share."""
    return IMPL.share_access_create(context, values)


def share_access_get(context, access_id):
    """Get share access rule."""
    return IMPL.share_access_get(context, access_id)


def share_access_get_all_for_share(context, share_id, filters=None):
    """Get all access rules for given share."""
    return IMPL.share_access_get_all_for_share(context, share_id,
                                               filters=filters)


def share_access_get_all_for_instance(context, instance_id, filters=None,
                                      with_share_access_data=True):
    """Get all access rules related to a certain share instance."""
    return IMPL.share_access_get_all_for_instance(
        context, instance_id, filters=filters,
        with_share_access_data=with_share_access_data)


def share_access_get_all_by_type_and_access(context, share_id, access_type,
                                            access):
    """Returns share access by given type and access."""
    return IMPL.share_access_get_all_by_type_and_access(
        context, share_id, access_type, access)


def share_access_check_for_existing_access(context, share_id, access_type,
                                           access_to):
    """Returns True if rule corresponding to the type and client exists."""
    return IMPL.share_access_check_for_existing_access(
        context, share_id, access_type, access_to)


def share_instance_access_create(context, values, share_instance_id):
    """Allow access to share instance."""
    return IMPL.share_instance_access_create(
        context, values, share_instance_id)


def share_instance_access_copy(context, share_id, instance_id):
    """Maps the existing access rules for the share to the instance in the DB.

    Adds the instance mapping to the share's access rules and
    returns the share's access rules.
    """
    return IMPL.share_instance_access_copy(context, share_id, instance_id)


def share_instance_access_get(context, access_id, instance_id,
                              with_share_access_data=True):
    """Get access rule mapping for share instance."""
    return IMPL.share_instance_access_get(
        context, access_id, instance_id,
        with_share_access_data=with_share_access_data)


def share_instance_access_update(context, access_id, instance_id, updates):
    """Update the access mapping row for a given share instance and access."""
    return IMPL.share_instance_access_update(
        context, access_id, instance_id, updates)


def share_instance_access_delete(context, mapping_id):
    """Deny access to share instance."""
    return IMPL.share_instance_access_delete(context, mapping_id)


def share_access_metadata_update(context, access_id, metadata):
    """Update metadata of share access rule."""
    return IMPL.share_access_metadata_update(context, access_id, metadata)


def share_access_metadata_delete(context, access_id, key):
    """Delete metadata of share access rule."""
    return IMPL.share_access_metadata_delete(context, access_id, key)


####################


def share_snapshot_instance_update(context, instance_id, values):
    """Set the given properties on a share snapshot instance and update it.

    Raises NotFound if snapshot instance does not exist.
    """
    return IMPL.share_snapshot_instance_update(context, instance_id, values)


def share_snapshot_instance_create(context, snapshot_id, values):
    """Create a share snapshot instance for an existing snapshot."""
    return IMPL.share_snapshot_instance_create(
        context, snapshot_id, values)


def share_snapshot_instance_get(context, instance_id, with_share_data=False):
    """Get a snapshot instance or raise a NotFound exception."""
    return IMPL.share_snapshot_instance_get(
        context, instance_id, with_share_data=with_share_data)


def share_snapshot_instance_get_all_with_filters(context, filters,
                                                 with_share_data=False):
    """Get all snapshot instances satisfying provided filters."""
    return IMPL.share_snapshot_instance_get_all_with_filters(
        context, filters, with_share_data=with_share_data)


def share_snapshot_instance_delete(context, snapshot_instance_id):
    """Delete a share snapshot instance."""
    return IMPL.share_snapshot_instance_delete(context, snapshot_instance_id)


####################


def share_snapshot_create(context, values):
    """Create a snapshot from the values dictionary."""
    return IMPL.share_snapshot_create(context, values)


def share_snapshot_get(context, snapshot_id):
    """Get a snapshot or raise if it does not exist."""
    return IMPL.share_snapshot_get(context, snapshot_id)


def share_snapshot_get_all(context, filters=None, sort_key=None,
                           sort_dir=None):
    """Get all snapshots."""
    return IMPL.share_snapshot_get_all(
        context, filters=filters, sort_key=sort_key, sort_dir=sort_dir,
    )


def share_snapshot_get_all_by_project(context, project_id, filters=None,
                                      sort_key=None, sort_dir=None):
    """Get all snapshots belonging to a project."""
    return IMPL.share_snapshot_get_all_by_project(
        context, project_id, filters=filters, sort_key=sort_key,
        sort_dir=sort_dir,
    )


def share_snapshot_get_all_for_share(context, share_id, filters=None,
                                     sort_key=None, sort_dir=None):
    """Get all snapshots for a share."""
    return IMPL.share_snapshot_get_all_for_share(
        context, share_id, filters=filters, sort_key=sort_key,
        sort_dir=sort_dir,
    )


def share_snapshot_get_latest_for_share(context, share_id):
    """Get the most recent snapshot for a share."""
    return IMPL.share_snapshot_get_latest_for_share(context, share_id)


def share_snapshot_update(context, snapshot_id, values):
    """Set the given properties on an snapshot and update it.

    Raises NotFound if snapshot does not exist.
    """
    return IMPL.share_snapshot_update(context, snapshot_id, values)


###################
def share_snapshot_access_create(context, values):
    """Create a share snapshot access from the values dictionary."""
    return IMPL.share_snapshot_access_create(context, values)


def share_snapshot_access_get(context, access_id):
    """Get share snapshot access rule from given access_id."""
    return IMPL.share_snapshot_access_get(context, access_id)


def share_snapshot_access_get_all_for_snapshot_instance(
        context, snapshot_instance_id, session=None):
    """Get all access rules related to a certain snapshot instance."""
    return IMPL.share_snapshot_access_get_all_for_snapshot_instance(
        context, snapshot_instance_id, session)


def share_snapshot_access_get_all_for_share_snapshot(context,
                                                     share_snapshot_id,
                                                     filters):
    """Get all access rules for a given share snapshot according to filters."""
    return IMPL.share_snapshot_access_get_all_for_share_snapshot(
        context, share_snapshot_id, filters)


def share_snapshot_check_for_existing_access(context, share_snapshot_id,
                                             access_type, access_to):
    """Returns True if rule corresponding to the type and client exists."""
    return IMPL.share_snapshot_check_for_existing_access(context,
                                                         share_snapshot_id,
                                                         access_type,
                                                         access_to)


def share_snapshot_export_locations_get(context, snapshot_id):
    """Get all export locations for a given share snapshot."""
    return IMPL.share_snapshot_export_locations_get(context, snapshot_id)


def share_snapshot_instance_access_update(
        context, access_id, instance_id, updates):
    """Update the state of the share snapshot instance access."""
    return IMPL.share_snapshot_instance_access_update(
        context, access_id, instance_id, updates)


def share_snapshot_instance_access_get(context, share_snapshot_instance_id,
                                       access_id):
    """Get the share snapshot instance access related to given ids."""
    return IMPL.share_snapshot_instance_access_get(
        context, share_snapshot_instance_id, access_id)


def share_snapshot_instance_access_delete(context, access_id,
                                          snapshot_instance_id):
    """Delete share snapshot instance access given its id."""
    return IMPL.share_snapshot_instance_access_delete(
        context, access_id, snapshot_instance_id)


def share_snapshot_instance_export_location_create(context, values):
    """Create a share snapshot instance export location."""
    return IMPL.share_snapshot_instance_export_location_create(context, values)


def share_snapshot_instance_export_locations_get_all(
        context, share_snapshot_instance_id):
    """Get the share snapshot instance export locations for given id."""
    return IMPL.share_snapshot_instance_export_locations_get_all(
        context, share_snapshot_instance_id)


def share_snapshot_instance_export_location_get(context, el_id):
    """Get the share snapshot instance export location for given id."""
    return IMPL.share_snapshot_instance_export_location_get(
        context, el_id)


def share_snapshot_instance_export_location_delete(context, el_id):
    """Delete share snapshot instance export location given its id."""
    return IMPL.share_snapshot_instance_export_location_delete(context, el_id)


###################
def security_service_create(context, values):
    """Create security service DB record."""
    return IMPL.security_service_create(context, values)


def security_service_delete(context, id):
    """Delete security service DB record."""
    return IMPL.security_service_delete(context, id)


def security_service_update(context, id, values):
    """Update security service DB record."""
    return IMPL.security_service_update(context, id, values)


def security_service_get(context, id):
    """Get security service DB record."""
    return IMPL.security_service_get(context, id)


def security_service_get_all(context):
    """Get all security service DB records."""
    return IMPL.security_service_get_all(context)


def security_service_get_all_by_project(context, project_id):
    """Get all security service DB records for the given project."""
    return IMPL.security_service_get_all_by_project(context, project_id)


####################


def share_metadata_get(context, share_id):
    """Get all metadata for a share."""
    return IMPL.share_metadata_get(context, share_id)


def share_metadata_delete(context, share_id, key):
    """Delete the given metadata item."""
    IMPL.share_metadata_delete(context, share_id, key)


def share_metadata_update(context, share, metadata, delete):
    """Update metadata if it exists, otherwise create it."""
    IMPL.share_metadata_update(context, share, metadata, delete)


###################

def share_export_location_get_by_uuid(context, export_location_uuid,
                                      ignore_secondary_replicas=False):
    """Get specific export location of a share."""
    return IMPL.share_export_location_get_by_uuid(
        context, export_location_uuid,
        ignore_secondary_replicas=ignore_secondary_replicas)


def share_export_locations_get(context, share_id):
    """Get all export locations of a share."""
    return IMPL.share_export_locations_get(context, share_id)


def share_export_locations_get_by_share_id(context, share_id,
                                           include_admin_only=True,
                                           ignore_migration_destination=False,
                                           ignore_secondary_replicas=False):
    """Get all export locations of a share by its ID."""
    return IMPL.share_export_locations_get_by_share_id(
        context, share_id, include_admin_only=include_admin_only,
        ignore_migration_destination=ignore_migration_destination,
        ignore_secondary_replicas=ignore_secondary_replicas)


def share_export_locations_get_by_share_instance_id(context,
                                                    share_instance_id,
                                                    include_admin_only=True):
    """Get all export locations of a share instance by its ID."""
    return IMPL.share_export_locations_get_by_share_instance_id(
        context, share_instance_id, include_admin_only=include_admin_only)


def share_export_locations_update(context, share_instance_id, export_locations,
                                  delete=True):
    """Update export locations of a share instance."""
    return IMPL.share_export_locations_update(
        context, share_instance_id, export_locations, delete)


####################

def export_location_metadata_get(context, export_location_uuid, session=None):
    """Get all metadata of an export location."""
    return IMPL.export_location_metadata_get(
        context, export_location_uuid, session=session)


def export_location_metadata_delete(context, export_location_uuid, keys,
                                    session=None):
    """Delete metadata of an export location."""
    return IMPL.export_location_metadata_delete(
        context, export_location_uuid, keys, session=session)


def export_location_metadata_update(context, export_location_uuid, metadata,
                                    delete, session=None):
    """Update metadata of an export location."""
    return IMPL.export_location_metadata_update(
        context, export_location_uuid, metadata, delete, session=session)

####################


def share_network_create(context, values):
    """Create a share network DB record."""
    return IMPL.share_network_create(context, values)


def share_network_delete(context, id):
    """Delete a share network DB record."""
    return IMPL.share_network_delete(context, id)


def share_network_update(context, id, values):
    """Update a share network DB record."""
    return IMPL.share_network_update(context, id, values)


def share_network_get(context, id):
    """Get requested share network DB record."""
    return IMPL.share_network_get(context, id)


def share_network_get_all(context):
    """Get all share network DB records."""
    return IMPL.share_network_get_all(context)


def share_network_get_all_by_project(context, project_id):
    """Get all share network DB records for the given project."""
    return IMPL.share_network_get_all_by_project(context, project_id)


def share_network_get_all_by_security_service(context, security_service_id):
    """Get all share network DB records for the given project."""
    return IMPL.share_network_get_all_by_security_service(
        context, security_service_id)


def share_network_add_security_service(context, id, security_service_id):
    return IMPL.share_network_add_security_service(context,
                                                   id,
                                                   security_service_id)


def share_network_remove_security_service(context, id, security_service_id):
    return IMPL.share_network_remove_security_service(context,
                                                      id,
                                                      security_service_id)


def count_share_networks(context, project_id, user_id=None,
                         share_type_id=None, session=None):
    return IMPL.count_share_networks(
        context, project_id, user_id=user_id, share_type_id=share_type_id,
        session=session,
    )

##################


def network_allocation_create(context, values):
    """Create a network allocation DB record."""
    return IMPL.network_allocation_create(context, values)


def network_allocation_delete(context, id):
    """Delete a network allocation DB record."""
    return IMPL.network_allocation_delete(context, id)


def network_allocation_update(context, id, values, read_deleted=None):
    """Update a network allocation DB record."""
    return IMPL.network_allocation_update(context, id, values,
                                          read_deleted=read_deleted)


def network_allocation_get(context, id, session=None, read_deleted=None):
    """Get a network allocation DB record."""
    return IMPL.network_allocation_get(context, id, session,
                                       read_deleted=read_deleted)


def network_allocations_get_for_share_server(context, share_server_id,
                                             session=None, label=None):
    """Get network allocations for share server."""
    return IMPL.network_allocations_get_for_share_server(
        context, share_server_id, label=label, session=session)


def network_allocations_get_by_ip_address(context, ip_address):
    """Get network allocations by IP address."""
    return IMPL.network_allocations_get_by_ip_address(context, ip_address)


##################


def share_server_create(context, values):
    """Create share server DB record."""
    return IMPL.share_server_create(context, values)


def share_server_delete(context, id):
    """Delete share server DB record."""
    return IMPL.share_server_delete(context, id)


def share_server_update(context, id, values):
    """Update share server DB record."""
    return IMPL.share_server_update(context, id, values)


def share_server_get(context, id, session=None):
    """Get share server DB record by ID."""
    return IMPL.share_server_get(context, id, session=session)


def share_server_search_by_identifier(context, identifier, session=None):
    """Search for share servers based on given identifier."""
    return IMPL.share_server_search_by_identifier(
        context, identifier, session=session)


def share_server_get_all_by_host_and_share_net_valid(context, host,
                                                     share_net_id,
                                                     session=None):
    """Get share server DB records by host and share net not error."""
    return IMPL.share_server_get_all_by_host_and_share_net_valid(
        context, host, share_net_id, session=session)


def share_server_get_all(context):
    """Get all share server DB records."""
    return IMPL.share_server_get_all(context)


def share_server_get_all_by_host(context, host):
    """Get all share servers related to particular host."""
    return IMPL.share_server_get_all_by_host(context, host)


def share_server_get_all_unused_deletable(context, host, updated_before):
    """Get all free share servers DB records."""
    return IMPL.share_server_get_all_unused_deletable(context, host,
                                                      updated_before)


def share_server_backend_details_set(context, share_server_id, server_details):
    """Create DB record with backend details."""
    return IMPL.share_server_backend_details_set(context, share_server_id,
                                                 server_details)


##################


def share_type_create(context, values, projects=None):
    """Create a new share type."""
    return IMPL.share_type_create(context, values, projects)


def share_type_get_all(context, inactive=False, filters=None):
    """Get all share types.

    :param context: context to query under
    :param inactive: Include inactive share types to the result set
    :param filters: Filters for the query in the form of key/value.
        :is_public: Filter share types based on visibility:

            * **True**: List public share types only
            * **False**: List private share types only
            * **None**: List both public and private share types

    :returns: list of matching share types
    """
    return IMPL.share_type_get_all(context, inactive, filters)


def share_type_get(context, type_id, inactive=False, expected_fields=None):
    """Get share type by id.

    :param context: context to query under
    :param type_id: share type id to get.
    :param inactive: Consider inactive share types when searching
    :param expected_fields: Return those additional fields.
                            Supported fields are: projects.
    :returns: share type
    """
    return IMPL.share_type_get(context, type_id, inactive, expected_fields)


def share_type_get_by_name(context, name):
    """Get share type by name."""
    return IMPL.share_type_get_by_name(context, name)


def share_type_get_by_name_or_id(context, name_or_id):
    """Get share type by name or ID and return None if not found."""
    return IMPL.share_type_get_by_name_or_id(context, name_or_id)


def share_type_access_get_all(context, type_id):
    """Get all share type access of a share type."""
    return IMPL.share_type_access_get_all(context, type_id)


def share_type_access_add(context, type_id, project_id):
    """Add share type access for project."""
    return IMPL.share_type_access_add(context, type_id, project_id)


def share_type_access_remove(context, type_id, project_id):
    """Remove share type access for project."""
    return IMPL.share_type_access_remove(context, type_id, project_id)


def share_type_destroy(context, id):
    """Delete a share type."""
    return IMPL.share_type_destroy(context, id)


####################


def share_type_extra_specs_get(context, share_type_id):
    """Get all extra specs for a share type."""
    return IMPL.share_type_extra_specs_get(context, share_type_id)


def share_type_extra_specs_delete(context, share_type_id, key):
    """Delete the given extra specs item."""
    return IMPL.share_type_extra_specs_delete(context, share_type_id, key)


def share_type_extra_specs_update_or_create(context, share_type_id,
                                            extra_specs):
    """Create or update share type extra specs.

    This adds or modifies the key/value pairs specified in the extra
    specs dict argument.
    """
    return IMPL.share_type_extra_specs_update_or_create(context,
                                                        share_type_id,
                                                        extra_specs)


def driver_private_data_get(context, entity_id, key=None, default=None):
    """Get one, list or all key-value pairs for given entity_id."""
    return IMPL.driver_private_data_get(context, entity_id, key, default)


def driver_private_data_update(context, entity_id, details,
                               delete_existing=False):
    """Update key-value pairs for given entity_id."""
    return IMPL.driver_private_data_update(context, entity_id, details,
                                           delete_existing)


def driver_private_data_delete(context, entity_id, key=None):
    """Remove one, list or all key-value pairs for given entity_id."""
    return IMPL.driver_private_data_delete(context, entity_id, key)


####################

def availability_zone_get(context, id_or_name):
    """Get availability zone by name or id."""
    return IMPL.availability_zone_get(context, id_or_name)


def availability_zone_get_all(context):
    """Get all active availability zones."""
    return IMPL.availability_zone_get_all(context)


####################

def share_group_get(context, share_group_id):
    """Get a share group or raise if it does not exist."""
    return IMPL.share_group_get(context, share_group_id)


def share_group_get_all(context, detailed=True, filters=None, sort_key=None,
                        sort_dir=None):
    """Get all share groups."""
    return IMPL.share_group_get_all(
        context, detailed=detailed, filters=filters, sort_key=sort_key,
        sort_dir=sort_dir)


def share_group_get_all_by_host(context, host, detailed=True, filters=None,
                                sort_key=None, sort_dir=None):
    """Get all share groups belonging to a host."""
    return IMPL.share_group_get_all_by_host(
        context, host, detailed=detailed, filters=filters, sort_key=sort_key,
        sort_dir=sort_dir)


def share_group_create(context, values):
    """Create a share group from the values dictionary."""
    return IMPL.share_group_create(context, values)


def share_group_get_all_by_share_server(context, share_server_id,
                                        filters=None, sort_key=None,
                                        sort_dir=None):
    """Get all share groups associated with a share server."""
    return IMPL.share_group_get_all_by_share_server(
        context, share_server_id, filters=filters, sort_key=sort_key,
        sort_dir=sort_dir)


def share_group_get_all_by_project(context, project_id, detailed=True,
                                   filters=None, sort_key=None,
                                   sort_dir=None):
    """Get all share groups belonging to a project."""
    return IMPL.share_group_get_all_by_project(
        context, project_id, detailed=detailed, filters=filters,
        sort_key=sort_key, sort_dir=sort_dir)


def share_group_update(context, share_group_id, values):
    """Set the given properties on a share group and update it.

    Raises NotFound if share group does not exist.
    """
    return IMPL.share_group_update(context, share_group_id, values)


def share_group_destroy(context, share_group_id):
    """Destroy the share group or raise if it does not exist."""
    return IMPL.share_group_destroy(context, share_group_id)


def count_shares_in_share_group(context, share_group_id):
    """Returns the number of undeleted shares with the specified group."""
    return IMPL.count_shares_in_share_group(context, share_group_id)


def get_all_shares_by_share_group(context, share_group_id):
    return IMPL.get_all_shares_by_share_group(context, share_group_id)


def count_share_group_snapshots_in_share_group(context, share_group_id):
    """Returns the number of sg snapshots with the specified share group."""
    return IMPL.count_share_group_snapshots_in_share_group(
        context, share_group_id)


def count_share_groups_in_share_network(context, share_network_id,
                                        session=None):
    """Return the number of groups with the specified share network."""
    return IMPL.count_share_groups_in_share_network(context, share_network_id)


def count_share_group_snapshot_members_in_share(context, share_id,
                                                session=None):
    """Returns the number of group snapshot members linked to the share."""
    return IMPL.count_share_group_snapshot_members_in_share(context, share_id)


def share_group_snapshot_get(context, share_group_snapshot_id):
    """Get a share group snapshot."""
    return IMPL.share_group_snapshot_get(context, share_group_snapshot_id)


def share_group_snapshot_get_all(context, detailed=True, filters=None,
                                 sort_key=None, sort_dir=None):
    """Get all share group snapshots."""
    return IMPL.share_group_snapshot_get_all(
        context, detailed=detailed, filters=filters, sort_key=sort_key,
        sort_dir=sort_dir)


def share_group_snapshot_get_all_by_project(context, project_id, detailed=True,
                                            filters=None, sort_key=None,
                                            sort_dir=None):
    """Get all share group snapshots belonging to a project."""
    return IMPL.share_group_snapshot_get_all_by_project(
        context, project_id, detailed=detailed, filters=filters,
        sort_key=sort_key, sort_dir=sort_dir)


def share_group_snapshot_create(context, values):
    """Create a share group snapshot from the values dictionary."""
    return IMPL.share_group_snapshot_create(context, values)


def share_group_snapshot_update(context, share_group_snapshot_id, values):
    """Set the given properties on a share group snapshot and update it.

    Raises NotFound if share group snapshot does not exist.
    """
    return IMPL.share_group_snapshot_update(
        context, share_group_snapshot_id, values)


def share_group_snapshot_destroy(context, share_group_snapshot_id):
    """Destroy the share_group_snapshot or raise if it does not exist."""
    return IMPL.share_group_snapshot_destroy(context, share_group_snapshot_id)


def share_group_snapshot_members_get_all(context, share_group_snapshot_id):
    """Return the members of a share group snapshot."""
    return IMPL.share_group_snapshot_members_get_all(
        context, share_group_snapshot_id)


def share_group_snapshot_member_create(context, values):
    """Create a share group snapshot member from the values dictionary."""
    return IMPL.share_group_snapshot_member_create(context, values)


def share_group_snapshot_member_update(context, member_id, values):
    """Set the given properties on a share group snapshot member and update it.

    Raises NotFound if share_group_snapshot member does not exist.
    """
    return IMPL.share_group_snapshot_member_update(context, member_id, values)


####################

def share_replicas_get_all(context, with_share_server=False,
                           with_share_data=False):
    """Returns all share replicas regardless of share."""
    return IMPL.share_replicas_get_all(
        context, with_share_server=with_share_server,
        with_share_data=with_share_data)


def share_replicas_get_all_by_share(context, share_id, with_share_server=False,
                                    with_share_data=False):
    """Returns all share replicas for a given share."""
    return IMPL.share_replicas_get_all_by_share(
        context, share_id, with_share_server=with_share_server,
        with_share_data=with_share_data)


def share_replicas_get_available_active_replica(context, share_id,
                                                with_share_server=False,
                                                with_share_data=False):
    """Returns an active replica for a given share."""
    return IMPL.share_replicas_get_available_active_replica(
        context, share_id, with_share_server=with_share_server,
        with_share_data=with_share_data)


def share_replica_get(context, replica_id, with_share_server=False,
                      with_share_data=False):
    """Get share replica by id."""
    return IMPL.share_replica_get(
        context, replica_id, with_share_server=with_share_server,
        with_share_data=with_share_data)


def share_replica_update(context, share_replica_id, values,
                         with_share_data=False):
    """Updates a share replica with given values."""
    return IMPL.share_replica_update(context, share_replica_id, values,
                                     with_share_data=with_share_data)


def share_replica_delete(context, share_replica_id):
    """Deletes a share replica."""
    return IMPL.share_replica_delete(context, share_replica_id)


def purge_deleted_records(context, age_in_days):
    """Purge deleted rows older than given age from all tables

    :raises: InvalidParameterValue if age_in_days is incorrect.
    """
    return IMPL.purge_deleted_records(context, age_in_days=age_in_days)


####################


def share_group_type_create(context, values, projects=None):
    """Create a new share group type."""
    return IMPL.share_group_type_create(context, values, projects)


def share_group_type_get_all(context, inactive=False, filters=None):
    """Get all share group types.

    :param context: context to query under
    :param inactive: Include inactive share group types to the result set
    :param filters: Filters for the query in the form of key/value.
        :is_public: Filter share group types based on visibility:

            * **True**: List public group types only
            * **False**: List private group types only
            * **None**: List both public and private group types

    :returns: list of matching share group types
    """
    return IMPL.share_group_type_get_all(context, inactive, filters)


def share_group_type_get(context, type_id, inactive=False,
                         expected_fields=None):
    """Get share_group type by id.

    :param context: context to query under
    :param type_id: group type id to get.
    :param inactive: Consider inactive group types when searching
    :param expected_fields: Return those additional fields.
                            Supported fields are: projects.
    :returns: share group type
    """
    return IMPL.share_group_type_get(
        context, type_id, inactive, expected_fields)


def share_group_type_get_by_name(context, name):
    """Get share group type by name."""
    return IMPL.share_group_type_get_by_name(context, name)


def share_group_type_access_get_all(context, type_id):
    """Get all share group type access of a share group type."""
    return IMPL.share_group_type_access_get_all(context, type_id)


def share_group_type_access_add(context, type_id, project_id):
    """Add share group type access for project."""
    return IMPL.share_group_type_access_add(context, type_id, project_id)


def share_group_type_access_remove(context, type_id, project_id):
    """Remove share group type access for project."""
    return IMPL.share_group_type_access_remove(context, type_id, project_id)


def share_group_type_destroy(context, type_id):
    """Delete a share group type."""
    return IMPL.share_group_type_destroy(context, type_id)


def share_group_type_specs_get(context, type_id):
    """Get all group specs for a share group type."""
    return IMPL.share_group_type_specs_get(context, type_id)


def share_group_type_specs_delete(context, type_id, key):
    """Delete the given group specs item."""
    return IMPL.share_group_type_specs_delete(context, type_id, key)


def share_group_type_specs_update_or_create(context, type_id, group_specs):
    """Create or update share group type specs.

    This adds or modifies the key/value pairs specified in the group
    specs dict argument.
    """
    return IMPL.share_group_type_specs_update_or_create(
        context, type_id, group_specs)


####################


def message_get(context, message_id):
    """Return a message with the specified ID."""
    return IMPL.message_get(context, message_id)


def message_get_all(context, filters=None, sort_key=None, sort_dir=None):
    """Returns all messages with the project of the specified context."""
    return IMPL.message_get_all(context, filters=filters, sort_key=sort_key,
                                sort_dir=sort_dir)


def message_create(context, values):
    """Creates a new message with the specified values."""
    return IMPL.message_create(context, values)


def message_destroy(context, message_id):
    """Deletes message with the specified ID."""
    return IMPL.message_destroy(context, message_id)


def cleanup_expired_messages(context):
    """Soft delete expired messages"""
    return IMPL.cleanup_expired_messages(context)


def backend_info_get(context, host):
    """Get hash info for given host."""
    return IMPL.backend_info_get(context, host)


def backend_info_update(context, host, value=None,
                        delete_existing=False):
    """Update hash info for host."""
    return IMPL.backend_info_update(context, host=host, value=value,
                                    delete_existing=delete_existing)



# Copyright (c) 2016 Alex Meade.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
NetApp Data ONTAP data motion library.

This library handles transferring data from a source to a destination. Its
responsibility is to handle this as efficiently as possible given the
location of the data's source and destination. This includes cloning,
SnapMirror, and copy-offload as improvements to brute force data transfer.
"""

from oslo_config import cfg
from oslo_log import log
from oslo_utils import excutils

from manila import exception
from manila.i18n import _
from manila.share import configuration
from manila.share import driver
from manila.share.drivers.netapp.dataontap.client import api as netapp_api
from manila.share.drivers.netapp.dataontap.client import client_cmode
from manila.share.drivers.netapp import options as na_opts
from manila.share.drivers.netapp import utils as na_utils
from manila.share import utils as share_utils
from manila import utils


LOG = log.getLogger(__name__)
CONF = cfg.CONF


def get_backend_configuration(backend_name):
    config_stanzas = CONF.list_all_sections()
    if backend_name not in config_stanzas:
        msg = _("Could not find backend stanza %(backend_name)s in "
                "configuration which is required for replication or migration "
                "workflows with the source backend. Available stanzas are "
                "%(stanzas)s")
        params = {
            "stanzas": config_stanzas,
            "backend_name": backend_name,
        }
        raise exception.BadConfigurationException(reason=msg % params)

    config = configuration.Configuration(driver.share_opts,
                                         config_group=backend_name)
    config.append_config_values(na_opts.netapp_cluster_opts)
    config.append_config_values(na_opts.netapp_connection_opts)
    config.append_config_values(na_opts.netapp_basicauth_opts)
    config.append_config_values(na_opts.netapp_transport_opts)
    config.append_config_values(na_opts.netapp_support_opts)
    config.append_config_values(na_opts.netapp_provisioning_opts)
    config.append_config_values(na_opts.netapp_data_motion_opts)

    return config


def get_client_for_backend(backend_name, vserver_name=None):
    config = get_backend_configuration(backend_name)
    client = client_cmode.NetAppCmodeClient(
        transport_type=config.netapp_transport_type,
        username=config.netapp_login,
        password=config.netapp_password,
        hostname=config.netapp_server_hostname,
        port=config.netapp_server_port,
        vserver=vserver_name or config.netapp_vserver,
        trace=na_utils.TRACE_API)

    return client


class DataMotionSession(object):

    def _get_backend_volume_name(self, config, share_obj):
        """Return the calculated backend name of the share.

        Uses the netapp_volume_name_template configuration value for the
        backend to calculate the volume name on the array for the share.
        """
        volume_name = config.netapp_volume_name_template % {
            'share_id': share_obj['id'].replace('-', '_')}
        return volume_name

    def _get_backend_qos_policy_group_name(self, share):
        """Get QoS policy name according to QoS policy group name template."""
        __, config = self._get_backend_config_obj(share)
        return config.netapp_qos_policy_group_name_template % {
            'share_id': share['id'].replace('-', '_')}

    def get_vserver_from_share(self, share_obj):
        share_server = share_obj.get('share_server')
        if share_server:
            backend_details = share_server.get('backend_details')
            if backend_details:
                return backend_details.get('vserver_name')

    def _get_backend_config_obj(self, share_obj):
        backend_name = share_utils.extract_host(
            share_obj['host'], level='backend_name')
        config = get_backend_configuration(backend_name)
        return backend_name, config

    def get_backend_info_for_share(self, share_obj):
        backend_name, config = self._get_backend_config_obj(share_obj)
        vserver = (self.get_vserver_from_share(share_obj) or
                   config.netapp_vserver)
        volume_name = self._get_backend_volume_name(
            config, share_obj)

        return volume_name, vserver, backend_name

    def get_snapmirrors(self, source_share_obj, dest_share_obj):
        dest_volume_name, dest_vserver, dest_backend = (
            self.get_backend_info_for_share(dest_share_obj))
        dest_client = get_client_for_backend(dest_backend,
                                             vserver_name=dest_vserver)

        src_volume_name, src_vserver, __ = self.get_backend_info_for_share(
            source_share_obj)

        snapmirrors = dest_client.get_snapmirrors(
            src_vserver, src_volume_name,
            dest_vserver, dest_volume_name,
            desired_attributes=['relationship-status',
                                'mirror-state',
                                'source-vserver',
                                'source-volume',
                                'last-transfer-end-timestamp'])
        return snapmirrors

    def create_snapmirror(self, source_share_obj, dest_share_obj):
        """Sets up a SnapMirror relationship between two volumes.

        1. Create SnapMirror relationship
        2. Initialize data transfer asynchronously
        """
        dest_volume_name, dest_vserver, dest_backend = (
            self.get_backend_info_for_share(dest_share_obj))
        dest_client = get_client_for_backend(dest_backend,
                                             vserver_name=dest_vserver)

        src_volume_name, src_vserver, __ = self.get_backend_info_for_share(
            source_share_obj)

        # 1. Create SnapMirror relationship
        # TODO(ameade): Change the schedule from hourly to a config value
        dest_client.create_snapmirror(src_vserver,
                                      src_volume_name,
                                      dest_vserver,
                                      dest_volume_name,
                                      schedule='hourly')

        # 2. Initialize async transfer of the initial data
        dest_client.initialize_snapmirror(src_vserver,
                                          src_volume_name,
                                          dest_vserver,
                                          dest_volume_name)

    def delete_snapmirror(self, source_share_obj, dest_share_obj,
                          release=True):
        """Ensures all information about a SnapMirror relationship is removed.

        1. Abort snapmirror
        2. Delete the snapmirror
        3. Release snapmirror to cleanup snapmirror metadata and snapshots
        """
        dest_volume_name, dest_vserver, dest_backend = (
            self.get_backend_info_for_share(dest_share_obj))
        dest_client = get_client_for_backend(dest_backend,
                                             vserver_name=dest_vserver)

        src_volume_name, src_vserver, src_backend = (
            self.get_backend_info_for_share(source_share_obj))

        # 1. Abort any ongoing transfers
        try:
            dest_client.abort_snapmirror(src_vserver,
                                         src_volume_name,
                                         dest_vserver,
                                         dest_volume_name,
                                         clear_checkpoint=False)
        except netapp_api.NaApiError as e:
            # Snapmirror is already deleted
            pass

        # 2. Delete SnapMirror Relationship and cleanup destination snapshots
        try:
            dest_client.delete_snapmirror(src_vserver,
                                          src_volume_name,
                                          dest_vserver,
                                          dest_volume_name)
        except netapp_api.NaApiError as e:
            with excutils.save_and_reraise_exception() as exc_context:
                if (e.code == netapp_api.EOBJECTNOTFOUND or
                        e.code == netapp_api.ESOURCE_IS_DIFFERENT or
                        "(entry doesn't exist)" in e.message):
                    LOG.info('No snapmirror relationship to delete')
                    exc_context.reraise = False

        if release:
            # If the source is unreachable, do not perform the release
            try:
                src_client = get_client_for_backend(src_backend,
                                                    vserver_name=src_vserver)
            except Exception:
                src_client = None
            # 3. Cleanup SnapMirror relationship on source
            try:
                if src_client:
                    src_client.release_snapmirror(src_vserver,
                                                  src_volume_name,
                                                  dest_vserver,
                                                  dest_volume_name)
            except netapp_api.NaApiError as e:
                with excutils.save_and_reraise_exception() as exc_context:
                    if (e.code == netapp_api.EOBJECTNOTFOUND or
                            e.code == netapp_api.ESOURCE_IS_DIFFERENT or
                            "(entry doesn't exist)" in e.message):
                        # Handle the case where the snapmirror is already
                        # cleaned up
                        exc_context.reraise = False

    def update_snapmirror(self, source_share_obj, dest_share_obj):
        """Schedule a snapmirror update to happen on the backend."""
        dest_volume_name, dest_vserver, dest_backend = (
            self.get_backend_info_for_share(dest_share_obj))
        dest_client = get_client_for_backend(dest_backend,
                                             vserver_name=dest_vserver)

        src_volume_name, src_vserver, __ = self.get_backend_info_for_share(
            source_share_obj)

        # Update SnapMirror
        dest_client.update_snapmirror(src_vserver,
                                      src_volume_name,
                                      dest_vserver,
                                      dest_volume_name)

    def quiesce_then_abort(self, source_share_obj, dest_share_obj):
        dest_volume_name, dest_vserver, dest_backend = (
            self.get_backend_info_for_share(dest_share_obj))
        dest_client = get_client_for_backend(dest_backend,
                                             vserver_name=dest_vserver)

        src_volume_name, src_vserver, __ = self.get_backend_info_for_share(
            source_share_obj)

        # 1. Attempt to quiesce, then abort
        dest_client.quiesce_snapmirror(src_vserver,
                                       src_volume_name,
                                       dest_vserver,
                                       dest_volume_name)

        config = get_backend_configuration(share_utils.extract_host(
            source_share_obj['host'], level='backend_name'))
        retries = config.netapp_snapmirror_quiesce_timeout / 5

        @utils.retry(exception.ReplicationException, interval=5,
                     retries=retries, backoff_rate=1)
        def wait_for_quiesced():
            snapmirror = dest_client.get_snapmirrors(
                src_vserver, src_volume_name, dest_vserver,
                dest_volume_name, desired_attributes=['relationship-status',
                                                      'mirror-state']
            )[0]
            if snapmirror.get('relationship-status') != 'quiesced':
                raise exception.ReplicationException(
                    reason=("Snapmirror relationship is not quiesced."))

        try:
            wait_for_quiesced()
        except exception.ReplicationException:
            dest_client.abort_snapmirror(src_vserver,
                                         src_volume_name,
                                         dest_vserver,
                                         dest_volume_name,
                                         clear_checkpoint=False)

    def break_snapmirror(self, source_share_obj, dest_share_obj, mount=True):
        """Breaks SnapMirror relationship.

        1. Quiesce any ongoing snapmirror transfers
        2. Wait until snapmirror finishes transfers and enters quiesced state
        3. Break snapmirror
        4. Mount the destination volume so it is exported as a share
        """
        dest_volume_name, dest_vserver, dest_backend = (
            self.get_backend_info_for_share(dest_share_obj))
        dest_client = get_client_for_backend(dest_backend,
                                             vserver_name=dest_vserver)

        src_volume_name, src_vserver, __ = self.get_backend_info_for_share(
            source_share_obj)

        # 1. Attempt to quiesce, then abort
        self.quiesce_then_abort(source_share_obj, dest_share_obj)

        # 2. Break SnapMirror
        dest_client.break_snapmirror(src_vserver,
                                     src_volume_name,
                                     dest_vserver,
                                     dest_volume_name)

        # 3. Mount the destination volume and create a junction path
        if mount:
            dest_client.mount_volume(dest_volume_name)

    def resync_snapmirror(self, source_share_obj, dest_share_obj):
        """Resync SnapMirror relationship. """
        dest_volume_name, dest_vserver, dest_backend = (
            self.get_backend_info_for_share(dest_share_obj))
        dest_client = get_client_for_backend(dest_backend,
                                             vserver_name=dest_vserver)

        src_volume_name, src_vserver, __ = self.get_backend_info_for_share(
            source_share_obj)

        dest_client.resync_snapmirror(src_vserver,
                                      src_volume_name,
                                      dest_vserver,
                                      dest_volume_name)

    def resume_snapmirror(self, source_share_obj, dest_share_obj):
        """Resume SnapMirror relationship from a quiesced state."""
        dest_volume_name, dest_vserver, dest_backend = (
            self.get_backend_info_for_share(dest_share_obj))
        dest_client = get_client_for_backend(dest_backend,
                                             vserver_name=dest_vserver)

        src_volume_name, src_vserver, __ = self.get_backend_info_for_share(
            source_share_obj)

        dest_client.resume_snapmirror(src_vserver,
                                      src_volume_name,
                                      dest_vserver,
                                      dest_volume_name)

    def change_snapmirror_source(self, replica,
                                 orig_source_replica,
                                 new_source_replica, replica_list):
        """Creates SnapMirror relationship from the new source to destination.

        1. Delete all snapmirrors involving the replica, but maintain
        snapmirror metadata and snapshots for efficiency
        2. Ensure a new source -> replica snapmirror exists
        3. Resync new source -> replica snapmirror relationship
        """

        replica_volume_name, replica_vserver, replica_backend = (
            self.get_backend_info_for_share(replica))
        replica_client = get_client_for_backend(replica_backend,
                                                vserver_name=replica_vserver)

        new_src_volume_name, new_src_vserver, __ = (
            self.get_backend_info_for_share(new_source_replica))

        # 1. delete
        for other_replica in replica_list:
            if other_replica['id'] == replica['id']:
                continue

            # We need to delete ALL snapmirror relationships
            # involving this replica but do not remove snapmirror metadata
            # so that the new snapmirror relationship is efficient.
            self.delete_snapmirror(other_replica, replica, release=False)
            self.delete_snapmirror(replica, other_replica, release=False)

        # 2. create
        # TODO(ameade): Update the schedule if needed.
        replica_client.create_snapmirror(new_src_vserver,
                                         new_src_volume_name,
                                         replica_vserver,
                                         replica_volume_name,
                                         schedule='hourly')
        # 3. resync
        replica_client.resync_snapmirror(new_src_vserver,
                                         new_src_volume_name,
                                         replica_vserver,
                                         replica_volume_name)

    @na_utils.trace
    def remove_qos_on_old_active_replica(self, orig_active_replica):
        old_active_replica_qos_policy = (
            self._get_backend_qos_policy_group_name(orig_active_replica)
        )
        replica_volume_name, replica_vserver, replica_backend = (
            self.get_backend_info_for_share(orig_active_replica))
        replica_client = get_client_for_backend(
            replica_backend, vserver_name=replica_vserver)
        try:
            replica_client.set_qos_policy_group_for_volume(
                replica_volume_name, 'none')
            replica_client.mark_qos_policy_group_for_deletion(
                old_active_replica_qos_policy)
        except exception.StorageCommunicationException:
            LOG.exception("Could not communicate with the backend "
                          "for replica %s to unset QoS policy and mark "
                          "the QoS policy group for deletion.",
                          orig_active_replica['id'])



# Copyright 2012 IBM Corp.
# Copyright 2014 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import datetime

import ddt
import mock
from oslo_utils import timeutils

from manila.api.v2 import services
from manila import context
from manila import db
from manila import exception
from manila import policy
from manila import test
from manila.tests.api import fakes


fake_services_list = [
    {
        'binary': 'manila-scheduler',
        'host': 'host1',
        'availability_zone': {'name': 'manila1'},
        'id': 1,
        'disabled': True,
        'updated_at': datetime.datetime(2012, 10, 29, 13, 42, 2),
        'created_at': datetime.datetime(2012, 9, 18, 2, 46, 27),
    },
    {
        'binary': 'manila-share',
        'host': 'host1',
        'availability_zone': {'name': 'manila1'},
        'id': 2,
        'disabled': True,
        'updated_at': datetime.datetime(2012, 10, 29, 13, 42, 5),
        'created_at': datetime.datetime(2012, 9, 18, 2, 46, 27)},
    {
        'binary': 'manila-scheduler',
        'host': 'host2',
        'availability_zone': {'name': 'manila2'},
        'id': 3,
        'disabled': False,
        'updated_at': datetime.datetime(2012, 9, 19, 6, 55, 34),
        'created_at': datetime.datetime(2012, 9, 18, 2, 46, 28)},
    {
        'binary': 'manila-share',
        'host': 'host2',
        'availability_zone': {'name': 'manila2'},
        'id': 4,
        'disabled': True,
        'updated_at': datetime.datetime(2012, 9, 18, 8, 3, 38),
        'created_at': datetime.datetime(2012, 9, 18, 2, 46, 28),
    },
]


fake_response_service_list = {'services': [
    {
        'id': 1,
        'binary': 'manila-scheduler',
        'host': 'host1',
        'zone': 'manila1',
        'status': 'disabled', 'state': 'up',
        'updated_at': datetime.datetime(2012, 10, 29, 13, 42, 2),
    },
    {
        'id': 2,
        'binary': 'manila-share',
        'host': 'host1',
        'zone': 'manila1',
        'status': 'disabled',
        'state': 'up',
        'updated_at': datetime.datetime(2012, 10, 29, 13, 42, 5),
    },
    {
        'id': 3,
        'binary': 'manila-scheduler',
        'host': 'host2',
        'zone': 'manila2',
        'status': 'enabled',
        'state': 'down',
        'updated_at': datetime.datetime(2012, 9, 19, 6, 55, 34),
    },
    {
        'id': 4,
        'binary': 'manila-share',
        'host': 'host2',
        'zone': 'manila2',
        'status': 'disabled',
        'state': 'down',
        'updated_at': datetime.datetime(2012, 9, 18, 8, 3, 38),
    },
]}


def fake_service_get_all(context):
    return fake_services_list


def fake_service_get_by_host_binary(context, host, binary):
    for service in fake_services_list:
        if service['host'] == host and service['binary'] == binary:
            return service
    return None


def fake_service_get_by_id(value):
    for service in fake_services_list:
        if service['id'] == value:
            return service
    return None


def fake_service_update(context, service_id, values):
    service = fake_service_get_by_id(service_id)
    if service is None:
        raise exception.ServiceNotFound(service_id=service_id)
    else:
        {'host': 'host1', 'binary': 'manila-share',
         'disabled': values['disabled']}


def fake_utcnow():
    return datetime.datetime(2012, 10, 29, 13, 42, 11)


@ddt.ddt
class ServicesTest(test.TestCase):

    def setUp(self):
        super(ServicesTest, self).setUp()

        self.mock_object(db, "service_get_all", fake_service_get_all)
        self.mock_object(timeutils, "utcnow", fake_utcnow)
        self.mock_object(db, "service_get_by_args",
                         fake_service_get_by_host_binary)
        self.mock_object(db, "service_update", fake_service_update)
        self.context = context.get_admin_context()
        self.controller = services.ServiceController()
        self.controller_legacy = services.ServiceControllerLegacy()
        self.resource_name = self.controller.resource_name
        self.mock_policy_check = self.mock_object(
            policy, 'check_policy', mock.Mock(return_value=True))

    @ddt.data(
        ('os-services', '1.0', services.ServiceControllerLegacy),
        ('os-services', '2.6', services.ServiceControllerLegacy),
        ('services', '2.7', services.ServiceController),
    )
    @ddt.unpack
    def test_services_list(self, url, version, controller):
        req = fakes.HTTPRequest.blank('/%s' % url, version=version)
        req.environ['manila.context'] = self.context

        res_dict = controller().index(req)

        self.assertEqual(fake_response_service_list, res_dict)
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'index')

    def test_services_list_with_host(self):
        req = fakes.HTTPRequest.blank('/services?host=host1', version='2.7')
        req.environ['manila.context'] = self.context

        res_dict = self.controller.index(req)

        response = {'services': [
            fake_response_service_list['services'][0],
            fake_response_service_list['services'][1],
        ]}
        self.assertEqual(response, res_dict)
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'index')

    def test_services_list_with_binary(self):
        req = fakes.HTTPRequest.blank(
            '/services?binary=manila-share', version='2.7')
        req.environ['manila.context'] = self.context

        res_dict = self.controller.index(req)

        response = {'services': [
            fake_response_service_list['services'][1],
            fake_response_service_list['services'][3],
        ]}

        self.assertEqual(response, res_dict)
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'index')

    def test_services_list_with_zone(self):
        req = fakes.HTTPRequest.blank('/services?zone=manila1', version='2.7')
        req.environ['manila.context'] = self.context

        res_dict = self.controller.index(req)

        response = {'services': [
            fake_response_service_list['services'][0],
            fake_response_service_list['services'][1],
        ]}
        self.assertEqual(response, res_dict)
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'index')

    def test_services_list_with_status(self):
        req = fakes.HTTPRequest.blank(
            '/services?status=enabled', version='2.7')
        req.environ['manila.context'] = self.context

        res_dict = self.controller.index(req)

        response = {'services': [
            fake_response_service_list['services'][2],
        ]}
        self.assertEqual(response, res_dict)
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'index')

    def test_services_list_with_state(self):
        req = fakes.HTTPRequest.blank('/services?state=up', version='2.7')
        req.environ['manila.context'] = self.context

        res_dict = self.controller.index(req)

        response = {'services': [
            fake_response_service_list['services'][0],
            fake_response_service_list['services'][1],
        ]}
        self.assertEqual(response, res_dict)
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'index')

    def test_services_list_with_host_binary(self):
        req = fakes.HTTPRequest.blank(
            "/services?binary=manila-share&state=up", version='2.7')
        req.environ['manila.context'] = self.context

        res_dict = self.controller.index(req)

        response = {'services': [fake_response_service_list['services'][1], ]}
        self.assertEqual(response, res_dict)
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'index')

    @ddt.data(
        ('os-services', '1.0', services.ServiceControllerLegacy),
        ('os-services', '2.6', services.ServiceControllerLegacy),
        ('services', '2.7', services.ServiceController),
    )
    @ddt.unpack
    def test_services_enable(self, url, version, controller):
        body = {'host': 'host1', 'binary': 'manila-share'}
        req = fakes.HTTPRequest.blank('/fooproject/%s' % url, version=version)

        res_dict = controller().update(req, "enable", body)

        self.assertFalse(res_dict['disabled'])
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'update')

    @ddt.data(
        ('os-services', '1.0', services.ServiceControllerLegacy),
        ('os-services', '2.6', services.ServiceControllerLegacy),
        ('services', '2.7', services.ServiceController),
    )
    @ddt.unpack
    def test_services_disable(self, url, version, controller):
        req = fakes.HTTPRequest.blank(
            '/fooproject/%s/disable' % url, version=version)
        body = {'host': 'host1', 'binary': 'manila-share'}

        res_dict = controller().update(req, "disable", body)

        self.assertTrue(res_dict['disabled'])
        self.mock_policy_check.assert_called_once_with(
            req.environ['manila.context'], self.resource_name, 'update')

    @ddt.data(
        ('os-services', '2.7', services.ServiceControllerLegacy),
        ('services', '2.6', services.ServiceController),
        ('services', '1.0', services.ServiceController),
    )
    @ddt.unpack
    def test_services_update_legacy_url_2_dot_7_api_not_found(self, url,
                                                              version,
                                                              controller):
        req = fakes.HTTPRequest.blank(
            '/fooproject/%s/fake' % url, version=version)
        body = {'host': 'host1', 'binary': 'manila-share'}

        self.assertRaises(
            exception.VersionNotFoundForAPIMethod,
            controller().update,
            req, "disable", body,
        )

    @ddt.data(
        ('os-services', '2.7', services.ServiceControllerLegacy),
        ('services', '2.6', services.ServiceController),
        ('services', '1.0', services.ServiceController),
    )
    @ddt.unpack
    def test_services_list_api_not_found(self, url, version, controller):
        req = fakes.HTTPRequest.blank('/fooproject/%s' % url, version=version)

        self.assertRaises(
            exception.VersionNotFoundForAPIMethod, controller().index, req)



# Copyright (c) 2011 OpenStack, LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Policy Engine For Manila"""

import functools
import sys

from oslo_config import cfg
from oslo_log import log as logging
from oslo_policy import policy
from oslo_utils import excutils

from manila import exception
from manila import policies

CONF = cfg.CONF
LOG = logging.getLogger(__name__)
_ENFORCER = None


def reset():
    global _ENFORCER
    if _ENFORCER:
        _ENFORCER.clear()
        _ENFORCER = None


def init(rules=None, use_conf=True):
    """Init an Enforcer class.

        :param policy_file: Custom policy file to use, if none is specified,
                          `CONF.policy_file` will be used.
        :param rules: Default dictionary / Rules to use. It will be
                    considered just in the first instantiation.
        :param default_rule: Default rule to use, CONF.default_rule will
                           be used if none is specified.
        :param use_conf: Whether to load rules from config file.
    """

    global _ENFORCER
    if not _ENFORCER:
        _ENFORCER = policy.Enforcer(CONF,
                                    rules=rules,
                                    use_conf=use_conf)
        register_rules(_ENFORCER)


def enforce(context, action, target, do_raise=True):
    """Verifies that the action is valid on the target in this context.

       :param context: manila context
       :param action: string representing the action to be checked,
           this should be colon separated for clarity.
           i.e. ``share:create``,
       :param target: dictionary representing the object of the action
           for object creation, this should be a dictionary representing the
           location of the object e.g. ``{'project_id': context.project_id}``
       :param do_raise: Whether to raise an exception if check fails.

       :returns: When ``do_raise`` is ``False``, returns a value that
                 evaluates as ``True`` or ``False`` depending on whether
                 the policy allows action on the target.

       :raises: manila.exception.PolicyNotAuthorized if verification fails
                and ``do_raise`` is ``True``.

    """
    init()
    if not isinstance(context, dict):
        context = context.to_dict()

    # Add the exception arguments if asked to do a raise
    extra = {}
    if do_raise:
        extra.update(exc=exception.PolicyNotAuthorized, action=action,
                     do_raise=do_raise)
    return _ENFORCER.enforce(action, target, context, **extra)


def set_rules(rules, overwrite=True, use_conf=False):
    """Set rules based on the provided dict of rules.

       :param rules: New rules to use. It should be an instance of dict.
       :param overwrite: Whether to overwrite current rules or update them
                         with the new rules.
       :param use_conf: Whether to reload rules from config file.
    """

    init(use_conf=False)
    _ENFORCER.set_rules(rules, overwrite, use_conf)


def get_rules():
    if _ENFORCER:
        return _ENFORCER.rules


def register_rules(enforcer):
    enforcer.register_defaults(policies.list_rules())


def get_enforcer():
    # This method is for use by oslopolicy CLI scripts. Those scripts need the
    # 'output-file' and 'namespace' options, but having those in sys.argv means
    # loading the Manila config options will fail as those are not expected to
    # be present. So we pass in an arg list with those stripped out.
    conf_args = []
    # Start at 1 because cfg.CONF expects the equivalent of sys.argv[1:]
    i = 1
    while i < len(sys.argv):
        if sys.argv[i].strip('-') in ['namespace', 'output-file']:
            i += 2
            continue
        conf_args.append(sys.argv[i])
        i += 1

    cfg.CONF(conf_args, project='manila')
    init()
    return _ENFORCER


def authorize(context, action, target, do_raise=True, exc=None):
    """Verifies that the action is valid on the target in this context.

       :param context: manila context
       :param action: string representing the action to be checked
           this should be colon separated for clarity.
           i.e. ``share:create``,
       :param target: dictionary representing the object of the action
           for object creation this should be a dictionary representing the
           location of the object e.g. ``{'project_id': context.project_id}``
       :param do_raise: if True (the default), raises PolicyNotAuthorized;
           if False, returns False
       :param exc: Class of the exception to raise if the check fails.
                   Any remaining arguments passed to :meth:`authorize` (both
                   positional and keyword arguments) will be passed to
                   the exception class. If not specified,
                   :class:`PolicyNotAuthorized` will be used.

       :raises manila.exception.PolicyNotAuthorized: if verification fails
           and do_raise is True. Or if 'exc' is specified it will raise an
           exception of that type.

       :return: returns a non-False value (not necessarily "True") if
           authorized, and the exact value False if not authorized and
           do_raise is False.
    """
    init()
    credentials = context.to_policy_values()
    if not exc:
        exc = exception.PolicyNotAuthorized
    try:
        result = _ENFORCER.authorize(action, target, credentials,
                                     do_raise=do_raise, exc=exc, action=action)
    except policy.PolicyNotRegistered:
        with excutils.save_and_reraise_exception():
            LOG.exception('Policy not registered')
    except Exception:
        with excutils.save_and_reraise_exception():
            LOG.debug('Policy check for %(action)s failed with credentials '
                      '%(credentials)s',
                      {'action': action, 'credentials': credentials})
    return result


def check_is_admin(context):
    """Whether or not user is admin according to policy setting.

    """
    init()

    credentials = context.to_policy_values()
    target = credentials
    return _ENFORCER.authorize('context_is_admin', target, credentials)


def wrap_check_policy(resource):
    """Check policy corresponding to the wrapped methods prior to execution."""
    def check_policy_wraper(func):
        @functools.wraps(func)
        def wrapped(self, context, target_obj, *args, **kwargs):
            check_policy(context, resource, func.__name__, target_obj)
            return func(self, context, target_obj, *args, **kwargs)

        return wrapped
    return check_policy_wraper


def check_policy(context, resource, action, target_obj=None, do_raise=True):
    target = {
        'project_id': context.project_id,
        'user_id': context.user_id,
    }
    target.update(target_obj or {})
    _action = '%s:%s' % (resource, action)
    return authorize(context, _action, target, do_raise=do_raise)



# Copyright 2015 Mirantis Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""
Tests For Pool Weigher.
"""

import mock
from oslo_config import cfg
from oslo_utils import timeutils

from manila import context
from manila.db import api as db_api
from manila.scheduler.weighers import base_host
from manila.scheduler.weighers import pool
from manila.share import utils
from manila import test
from manila.tests.scheduler import fakes

CONF = cfg.CONF


class PoolWeigherTestCase(test.TestCase):
    def setUp(self):
        super(PoolWeigherTestCase, self).setUp()
        self.host_manager = fakes.FakeHostManager()
        self.weight_handler = base_host.HostWeightHandler(
            'manila.scheduler.weighers')
        share_servers = [
            {'id': 'fake_server_id0'},
            {'id': 'fake_server_id1'},
            {'id': 'fake_server_id2'},
            {'id': 'fake_server_id3'},
            {'id': 'fake_server_id4'},
        ]
        services = [
            dict(id=1, host='host1@AAA', topic='share', disabled=False,
                 availability_zone='zone1', updated_at=timeutils.utcnow()),
            dict(id=2, host='host2@BBB', topic='share', disabled=False,
                 availability_zone='zone1', updated_at=timeutils.utcnow()),
            dict(id=3, host='host3@CCC', topic='share', disabled=False,
                 availability_zone='zone2', updated_at=timeutils.utcnow()),
            dict(id=4, host='host@DDD', topic='share', disabled=False,
                 availability_zone='zone3', updated_at=timeutils.utcnow()),
            dict(id=5, host='host5@EEE', topic='share', disabled=False,
                 availability_zone='zone3', updated_at=timeutils.utcnow()),
        ]
        self.host_manager.service_states = (
            fakes.SHARE_SERVICE_STATES_WITH_POOLS)
        self.mock_object(db_api, 'share_server_get_all_by_host',
                         mock.Mock(return_value=share_servers))
        self.mock_object(db_api.IMPL, 'service_get_all_by_topic',
                         mock.Mock(return_value=services))

    def _get_weighed_host(self, hosts, weight_properties=None):
        if weight_properties is None:
            weight_properties = {
                'server_pools_mapping': {
                    'fake_server_id2': [{'pool_name': 'pool2'}, ],
                },
            }
        return self.weight_handler.get_weighed_objects(
            [pool.PoolWeigher],
            hosts,
            weight_properties)[0]

    def _get_all_hosts(self):
        ctxt = context.get_admin_context()
        host_states = self.host_manager.get_all_host_states_share(ctxt)
        db_api.IMPL.service_get_all_by_topic.assert_called_once_with(
            ctxt, CONF.share_topic)
        return host_states

    def test_no_server_pool_mapping(self):
        weight_properties = {
            'server_pools_mapping': {},
        }
        weighed_host = self._get_weighed_host(self._get_all_hosts(),
                                              weight_properties)
        self.assertEqual(0.0, weighed_host.weight)

    def test_choose_pool_with_existing_share_server(self):
        # host1: weight = 0*(1.0)
        # host2: weight = 1*(1.0)
        # host3: weight = 0*(1.0)
        # host4: weight = 0*(1.0)
        # host5: weight = 0*(1.0)

        # so, host2 should win:

        weighed_host = self._get_weighed_host(self._get_all_hosts())
        self.assertEqual(1.0, weighed_host.weight)
        self.assertEqual(
            'host2@BBB', utils.extract_host(weighed_host.obj.host))

    def test_pool_weight_multiplier_positive(self):
        self.flags(pool_weight_multiplier=2.0)

        # host1: weight = 0*(2.0)
        # host2: weight = 1*(2.0)
        # host3: weight = 0*(2.0)
        # host4: weight = 0*(2.0)
        # host5: weight = 0*(2.0)

        # so, host2 should win:

        weighed_host = self._get_weighed_host(self._get_all_hosts())
        self.assertEqual(2.0, weighed_host.weight)
        self.assertEqual(
            'host2@BBB', utils.extract_host(weighed_host.obj.host))

    def test_pool_weight_multiplier_negative(self):
        self.flags(pool_weight_multiplier=-1.0)
        weight_properties = {
            'server_pools_mapping': {
                'fake_server_id0': [{'pool_name': 'pool1'}],
                'fake_server_id2': [{'pool_name': 'pool3'}],
                'fake_server_id3': [
                    {'pool_name': 'pool4a'},
                    {'pool_name': 'pool4b'},
                ],
                'fake_server_id4': [
                    {'pool_name': 'pool5a'},
                    {'pool_name': 'pool5b'},
                ],
            },
        }

        # host1: weight = 1*(-1.0)
        # host2: weight = 0*(-1.0)
        # host3: weight = 1*(-1.0)
        # host4: weight = 1*(-1.0)
        # host5: weight = 1*(-1.0)

        # so, host2 should win:
        weighed_host = self._get_weighed_host(self._get_all_hosts(),
                                              weight_properties)
        self.assertEqual(0.0, weighed_host.weight)
        self.assertEqual(
            'host2@BBB', utils.extract_host(weighed_host.obj.host))

    def test_pool_weigher_all_pools_with_share_servers(self):
        weight_properties = {
            'server_pools_mapping': {
                'fake_server_id0': [{'pool_name': 'pool1'}],
                'fake_server_id1': [{'pool_name': 'pool2'}],
                'fake_server_id2': [{'pool_name': 'pool3'}],
                'fake_server_id3': [
                    {'pool_name': 'pool4a'},
                    {'pool_name': 'pool4b'},
                ],
                'fake_server_id4': [
                    {'pool_name': 'pool5a'},
                    {'pool_name': 'pool5b'},
                ],
            },
        }

        # host1: weight = 1*(1.0)
        # host2: weight = 1*(1.0)
        # host3: weight = 1*(1.0)
        # host4: weight = 1*(1.0)
        # host5: weight = 1*(1.0)

        # But after normalization all weighers will be 0

        weighed_host = self._get_weighed_host(self._get_all_hosts(),
                                              weight_properties)
        self.assertEqual(0.0, weighed_host.weight)



# Copyright (c) 2013 OpenStack Foundation
# Copyright (c) 2015 Mirantis inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from manila.api.openstack import wsgi
from manila.api.views import availability_zones as availability_zones_views
from manila import db


class AvailabilityZoneMixin(object):
    """The Availability Zone API controller common logic.

    Mixin class that should be inherited by Availability Zone API controllers,
    which are used for different API URLs and microversions.
    """

    resource_name = "availability_zone"
    _view_builder_class = availability_zones_views.ViewBuilder

    @wsgi.Controller.authorize("index")
    def _index(self, req):
        """Describe all known availability zones."""
        views = db.availability_zone_get_all(req.environ['manila.context'])
        return self._view_builder.detail_list(views)


class AvailabilityZoneControllerLegacy(AvailabilityZoneMixin, wsgi.Controller):
    """Deprecated Availability Zone API controller.

    Used by legacy API v1 and v2 microversions from 2.0 to 2.6.
    Registered under deprecated API URL 'os-availability-zone'.
    """

    @wsgi.Controller.api_version('1.0', '2.6')
    def index(self, req):
        return self._index(req)


class AvailabilityZoneController(AvailabilityZoneMixin, wsgi.Controller):
    """Availability Zone API controller.

    Used only by API v2 starting from microversion 2.7.
    Registered under API URL 'availability-zones'.
    """

    @wsgi.Controller.api_version('2.7')
    def index(self, req):
        return self._index(req)


def create_resource_legacy():
    return wsgi.Resource(AvailabilityZoneControllerLegacy())


def create_resource():
    return wsgi.Resource(AvailabilityZoneController())
